from intake.source.base import DataSource


class FooPlugin(DataSource):
    name = 'otherfoo'
    version = '0.1'
    container = 'dataframe'
    partition_access = False

    def __init__(self, **kwargs):
        pass



import os
import numpy as np
import pytest
from ..npy import NPySource


@pytest.mark.parametrize('shape', [(1, ), (1, 1), (10, ), (5, 2), (3, 3, 3)])
def test_one_file(tempdir, shape):
    size = 1
    for s in shape:
        size *= s
    data = np.random.randint(1, 100, size=size).reshape(shape)
    fn = os.path.join(tempdir, 'out.npy')
    np.save(fn, data)
    s = NPySource(fn)
    out = s.read()
    assert (out == data).all()
    s = NPySource(fn, chunks=1)
    out = s.read()
    assert (out == data).all()
    s = NPySource(fn, shape=shape, dtype='int', chunks=1)
    out = s.read()
    assert (out == data).all()


@pytest.mark.parametrize('shape', [(1, ), (1, 1), (10, ), (5, 2), (3, 3, 3)])
def test_multi_file(tempdir, shape):
    size = 1
    for s in shape:
        size *= s
    data0 = np.random.randint(1, 100, size=size).reshape(shape)
    fn0 = os.path.join(tempdir, 'out0.npy')
    np.save(fn0, data0)
    data1 = np.random.randint(1, 100, size=size).reshape(shape)
    fn1 = os.path.join(tempdir, 'out1.npy')
    np.save(fn1, data1)
    data = np.stack([data0, data1])
    fn = [fn0, fn1]
    s = NPySource(fn)
    out = s.read()
    assert (out == data).all()
    s = NPySource(fn, chunks=1)
    out = s.read()
    assert (out == data).all()
    s = NPySource(fn, shape=shape, dtype='int', chunks=1)
    out = s.read()
    assert (out == data).all()
    s = NPySource(os.path.join(tempdir, 'out*.npy'))
    out = s.read()
    assert (out == data).all()



from intake.container.base import RemoteSource, get_partition
from intake.source.base import Schema


class RemoteSequenceSource(RemoteSource):
    """Sequence-of-things source on an Intake server"""
    name = 'remote_sequence'
    container = 'python'

    def __init__(self, url, headers, **kwargs):
        self.url = url
        self.npartitions = kwargs.get('npartition', 1)
        self.partition_access = self.npartitions > 1
        self.headers = headers
        self.metadata = kwargs.get('metadata', {})
        self._schema = Schema(npartitions=self.npartitions,
                              extra_metadata=self.metadata)
        self.bag = None
        super(RemoteSequenceSource, self).__init__(url, headers, **kwargs)

    def _load_metadata(self):
        import dask.bag as db
        import dask
        if self.bag is None:
            self.parts = [dask.delayed(get_partition)(
                self.url, self.headers, self._source_id, self.container, i
            )
                          for i in range(self.npartitions)]
            self.bag = db.from_delayed(self.parts)
        return self._schema

    def _get_partition(self, i):
        self._load_metadata()
        return self.parts[i].compute()

    def read(self):
        self._load_metadata()
        return self.bag.compute()

    def to_dask(self):
        self._load_metadata()
        return self.bag

    def _close(self):
        self.bag = None



import os
import subprocess
import time

import pytest
import requests

from intake.util_tests import ex, PY2
here = os.path.dirname(__file__)


MIN_PORT = 7480
MAX_PORT = 7489
PORT = MIN_PORT


def ping_server(url, swallow_exception, head=None):
    try:
        r = requests.get(url)
    except Exception as e:
        if swallow_exception:
            return False
        else:
            raise e

    return r.status_code in (200, 403)  # allow forbidden as well


def pick_port():
    global PORT
    port = PORT
    if port == MAX_PORT:
        PORT = MIN_PORT
    else:
        PORT += 1

    return port


@pytest.fixture(scope="module")
def intake_server(request):
    os.environ['INTAKE_DEBUG'] = 'true'
    # Catalog path comes from the test module
    path = request.module.TEST_CATALOG_PATH
    if isinstance(path, list):
        catalog_path = [p + '/*' for p in path]
    elif isinstance(path, str) and not path.endswith(
            '.yml') and not path.endswith('.yaml'):
        catalog_path = path + '/*'
    else:
        catalog_path = path
    server_conf = getattr(request.module, 'TEST_SERVER_CONF', None)

    # Start a catalog server on nonstandard port

    env = dict(os.environ)
    env['INTAKE_TEST'] = 'server'
    if server_conf is not None:
        env['INTAKE_CONF_FILE'] = server_conf
    port = pick_port()
    cmd = [ex, '-m', 'intake.cli.server', '--sys-exit-on-sigterm',
           '--port', str(port)]
    if isinstance(catalog_path, list):
        cmd.extend(catalog_path)
    else:
        cmd.append(catalog_path)
    try:
        p = subprocess.Popen(cmd, env=env)
        url = 'http://localhost:%d/v1/info' % port

        # wait for server to finish initalizing, but let the exception through
        # on last retry
        retries = 300
        try:
            while not ping_server(url, swallow_exception=(retries > 1)):
                time.sleep(0.1)
                retries -= 1
        except Exception:
            print(p.communicate())
            raise
        assert retries > 0, "Server never appeared"

        yield 'intake://localhost:%d' % port
    finally:
        if server_conf:
            try:
                env.pop('INTAKE_CONF_FILE', None)
                os.remove(server_conf)
            except:
                pass
        p.terminate()
        time.sleep(1)
        p.kill()


@pytest.fixture(scope='module')
def http_server():
    if PY2:
        cmd = ['python', '-m', 'SimpleHTTPServer', '8000']
    else:
        cmd = ['python', '-m', 'http.server', '8000']
    p = subprocess.Popen(cmd, cwd=os.path.join(here, 'catalog', 'tests'))
    timeout = 5
    while True:
        try:
            requests.get('http://localhost:8000/')
            break
        except:
            time.sleep(0.1)
            timeout -= 0.1
            assert timeout > 0, "timeout waiting for http server"
    try:
        yield 'http://localhost:8000/'
    finally:
        p.terminate()
        p.communicate()


@pytest.fixture()
def tempdir():
    import tempfile
    import shutil
    d = str(tempfile.mkdtemp())
    try:
        yield d
    finally:
        shutil.rmtree(d)



from .. import utils


def test_expand_templates():
    pars = {'a': "{{par}} hi"}
    context = {'b': 1, 'par': 'ho'}
    assert utils.expand_templates(pars, context)['a'] == 'ho hi'
    assert utils.expand_templates(pars, context, True)[1] == {'b'}


def test_expand_nested_template():
    pars = {'a': ["{{par}} hi"]}
    context = {'b': 1, 'par': 'ho'}
    assert utils.expand_templates(pars, context)['a'] == ['ho hi']
    assert utils.expand_templates(pars, context, True)[1] == {'b'}

    pars = {'a': {'k': {("{{par}} hi", )}}}
    context = {'b': 1, 'par': 'ho'}
    assert utils.expand_templates(pars, context)['a'] == {'k': {("ho hi", )}}
    assert utils.expand_templates(pars, context, True)[1] == {'b'}



import pkgutil
import warnings
import importlib
import inspect
import time
import logging

from .base import DataSource
from ..catalog.base import Catalog
logger = logging.getLogger('intake')


def autodiscover(path=None, plugin_prefix='intake_'):
    """Scan for Intake plugin packages and return a dict of plugins.

    This function searches path (or sys.path) for packages with names that
    start with plugin_prefix.  Those modules will be imported and scanned for
    subclasses of intake.source.base.Plugin.  Any subclasses found will be
    instantiated and returned in a dictionary, with the plugin's name attribute
    as the key.
    """

    plugins = {}

    for importer, name, ispkg in pkgutil.iter_modules(path=path):
        if name.startswith(plugin_prefix):
            t = time.time()
            new_plugins = load_plugins_from_module(name)

            for plugin_name, plugin in new_plugins.items():
                if plugin_name in plugins:
                    orig_path = inspect.getfile(plugins[plugin_name])
                    new_path = inspect.getfile(plugin)
                    warnings.warn('Plugin name collision for "%s" from'
                                  '\n    %s'
                                  '\nand'
                                  '\n    %s'
                                  '\nKeeping plugin from first location.'
                                  % (plugin_name, orig_path, new_path))
                else:
                    plugins[plugin_name] = plugin
            logger.debug("Import %s took: %7.2f s" % (name, time.time() - t))

    return plugins


def load_plugins_from_module(module_name):
    """Imports a module and returns dictionary of discovered Intake plugins.

    Plugin classes are instantiated and added to the dictionary, keyed by the
    name attribute of the plugin object.
    """
    plugins = {}

    try:
        if module_name.endswith('.py'):
            import imp
            mod = imp.load_source('module.name', module_name)
        else:
            mod = importlib.import_module(module_name)
    except Exception as e:
        logger.debug("Import module <{}> failed: {}".format(module_name, e))
        return {}
    for _, cls in inspect.getmembers(mod, inspect.isclass):
        # Don't try to register plugins imported into this module elsewhere
        if issubclass(cls, (Catalog, DataSource)):
            plugins[cls.name] = cls

    return plugins



import glob
import os
from intake.source.textfiles import TextFilesSource


def test_textfiles(tempdir):
    open(os.path.join(tempdir, '1.txt'), 'wt').write('hello\nworld')
    open(os.path.join(tempdir, '2.txt'), 'wt').write('hello\nworld')
    path = os.path.join(tempdir, '*.txt')
    t = TextFilesSource(path)
    t.discover()
    assert t.npartitions == len(glob.glob(path))
    assert t._get_partition(0) == t.to_dask().to_delayed()[0].compute()
    out = t.read()
    assert isinstance(out, list)
    assert out[0] == 'hello\n'



import os
import pytest
import shutil

from intake.source.cache import FileCache, CacheMetadata
import intake
import intake.config
from intake.source.tests.util import temp_cache
here = os.path.dirname(os.path.abspath(__file__))
import logging
logger = logging.getLogger('intake')
logging.basicConfig()


@pytest.fixture
def file_cache():
    return FileCache('csv', 
                     {'argkey': 'urlpath', 'regex': 'test/path', 'type': 'file'})


def test_ensure_cache_dir(file_cache):
    file_cache._ensure_cache_dir()
    assert os.path.exists(file_cache._cache_dir)

    file_cache.clear_all()
    shutil.rmtree(file_cache._cache_dir)

    with open(file_cache._cache_dir, 'w') as f:
        f.write('')
    
    with pytest.raises(Exception):
        file_cache._ensure_cache_dir()

    os.remove(file_cache._cache_dir)
    
    file_cache.clear_all()


def test_munge_path(file_cache):
    subdir = 'subdir'
    cache_path = file_cache._munge_path(subdir, 'test/path/foo.cvs')
    assert subdir in cache_path
    assert 'test/path' not in cache_path

    file_cache._spec['regex'] = 'https://example.com'
    cache_path = file_cache._munge_path(subdir, 'https://example.com/catalog.yml')
    assert subdir in cache_path
    assert file_cache._cache_dir in cache_path
    assert 'http' not in cache_path


def test_hash(file_cache):
    subdir = file_cache._hash('foo/bar.csv')

    import string
    # Checking for md5 hash
    assert all(c in string.hexdigits for c in subdir)

    file_cache._driver = 'bar'
    subdir_new = file_cache._hash('foo/bar.csv')
    assert subdir_new != subdir

    file_cache._driver = 'csv'
    subdir_new = file_cache._hash('foo/bar.csv')
    assert subdir_new == subdir

    file_cache._spec['regex'] = 'foo/bar'
    subdir_new = file_cache._hash('foo/bar.csv')
    assert subdir_new != subdir


def test_path(file_cache):
    urlpath = 'test/path/foo.csv'
    file_cache._spec['regex'] = 'test/path/'
    cache_path = file_cache._path(urlpath)

    assert file_cache._cache_dir in cache_path
    assert '//' not in cache_path[1:]
    file_cache.clear_all()


def test_path_no_match(file_cache):
    "No match should be a noop."
    urlpath = 'https://example.com/foo.csv'
    cache_path = file_cache._path(urlpath)
    assert urlpath == cache_path


def test_dir_cache(tempdir, temp_cache):
    [os.makedirs(os.path.join(tempdir, d)) for d in [
        'main', 'main/sub1', 'main/sub2']]
    for f in ['main/afile', 'main/sub1/subfile', 'main/sub2/subfile1',
              'main/sub2/subfile2']:
        fn = os.path.join(tempdir, f)
        with open(fn, 'w') as fo:
            fo.write(f)
    fn = os.path.join(tempdir, 'cached.yaml')
    shutil.copy2(os.path.join(here, 'cached.yaml'), fn)
    cat = intake.open_catalog(fn)
    s = cat.dirs()
    out = s.cache[0].load(s._urlpath, output=False)
    assert out[0] == os.path.join(tempdir, s.cache[0]._path(s._urlpath))
    assert open(os.path.join(out[0], 'afile')).read() == 'main/afile'
    md = CacheMetadata()
    got = md[s._urlpath]

    # Avoid re-copy
    s = cat.dirs()
    s.cache[0].load(s._urlpath, output=False)
    md2 = CacheMetadata()
    got2 = md2[s._urlpath]
    assert got == got2


def test_compressed_cache(temp_cache):
    cat = intake.open_catalog(os.path.join(here, 'cached.yaml'))
    s = cat.calvert()
    old = intake.config.conf['cache_download_progress']
    try:
        intake.config.conf['cache_download_progress'] = False
        df = s.read()
        assert len(df)
        md = CacheMetadata()
        assert len(md[s._urlpath]) == 1  # we gained exactly one CSV
        intake.config.conf['cache_download_progress'] = False
        df = s.read()
        assert len(df)
        md = CacheMetadata()
        assert len(md[s._urlpath]) == 1  # we still have exactly one CSV
    finally:
        intake.config.conf['cache_download_progress'] = old


def test_cache_to_cat(tempdir):
    old = intake.config.conf.copy()
    olddir = intake.config.confdir
    intake.config.confdir = str(tempdir)
    intake.config.conf.update({'cache_dir': 'catdir',
                               'cache_download_progress': False,
                               'cache_disabled': False})
    try:
        fn0 = os.path.join(here, 'calvert_uk.zip')
        fn1 = os.path.join(tempdir, 'calvert_uk.zip')
        shutil.copy2(fn0, fn1)
        fn0 = os.path.join(here, 'cached.yaml')
        fn1 = os.path.join(tempdir, 'cached.yaml')
        shutil.copy2(fn0, fn1)
        cat = intake.open_catalog(fn1)
        s = cat.calvert()
        df = s.read()
        assert len(df)
        md = CacheMetadata()
        f = md[s._urlpath][0]
        assert f['cache_path'].startswith(str(tempdir))
        assert 'intake_cache' in os.listdir(tempdir)
        assert os.listdir(os.path.join(tempdir, 'intake_cache'))
    finally:
        intake.config.confdir = olddir
        intake.config.conf.update(old)



def test_compressed_cache_infer(temp_cache):
    cat = intake.open_catalog(os.path.join(here, 'cached.yaml'))
    s = cat.calvert_infer()
    old = intake.config.conf['cache_download_progress']
    try:
        intake.config.conf['cache_download_progress'] = False
        df = s.read()
        assert len(df)
    finally:
        intake.config.conf['cache_download_progress'] = old


def test_compressed_cache_bad(temp_cache):
    cat = intake.open_catalog(os.path.join(here, 'cached.yaml'))
    s = cat.calvert_badkey()
    old = intake.config.conf['cache_download_progress']
    try:
        intake.config.conf['cache_download_progress'] = False
        with pytest.raises(ValueError):
            s.read()
    finally:
        intake.config.conf['cache_download_progress'] = old



import os
import sys

import pytest

from intake.source import discovery


@pytest.fixture
def extra_pythonpath():
    basedir = os.path.dirname(__file__)
    extra_path = os.path.join(basedir, 'plugin_searchpath')

    # Put extra directory on the python path
    sys.path.append(extra_path)

    yield extra_path

    # Return python path back to normal
    sys.path.remove(extra_path)


def test_discover(extra_pythonpath):
    registry = discovery.autodiscover()

    # possible other plugins in environment
    'foo' in registry
    registry['foo']()
    registry.pop('foo', None)


def test_discover_pluginprefix(extra_pythonpath):
    registry = discovery.autodiscover(plugin_prefix='not_intake_')

    assert 'otherfoo' in registry
    registry['otherfoo']()
    registry.pop('otherfoo', None)


def test_discover_collision(extra_pythonpath):
    with pytest.warns(UserWarning):
        registry = discovery.autodiscover(plugin_prefix='collision_')



class CatalogEntry(object):
    """A single item appearing in a catalog

    This is the base class, used by local entries (i.e., read from a YAML file)
    and by remote entries (read from a server).
    """
    def __init__(self, getenv=True, getshell=True):
        self._default_source = None
        self.getenv = getenv
        self.getshell = getshell

    def __repr__(self):
        return repr(self.describe())

    def describe(self):
        """Get a dictionary of attributes of this entry.

        Returns: dict with keys

          container : str 
              kind of container used by this data source
          description : str
              Markdown-friendly description of data source
          direct_access : str
              Mode of remote access: forbid, allow, force
          user_parameters : list[dict]
              List of user parameters defined by this entry

        """
        raise NotImplementedError

    def describe_open(self, **user_parameters):
        """Get a dictionary describing how to open this data source.

        Parameters
        ----------
          user_parameters : dict
              Values for user-configurable parameters for this data source

        Returns: dict with keys
          plugin : str
              Name of data plugin to use
          description : str
              Markdown-friendly description of data source 
          direct_access : str
              Mode of remote access: forbid, allow, force
          metadata : dict
              Dictionary of metadata defined in the catalog for this data source
          args: dict
              Dictionary of keyword arguments for the plugin
        
        """
        raise NotImplementedError

    def get(self, **user_parameters):
        """Open the data source.

        Equivalent to calling the catalog entry like a function.

        Parameters
        ----------
          user_parameters : dict
              Values for user-configurable parameters for this data source

        Returns: DataSource
        """
        raise NotImplementedError

    # Convenience methods for configuring source and automatically
    # creating a default source when needed.
    def __call__(self, **kwargs):
        return self.get(**kwargs)

    def _get_default_source(self):
        if self._default_source is None:
            self._default_source = self.get()
        return self._default_source

    @property
    def plots(self):
        """List custom associated quick-plots """
        return list(self._metadata.get('plots', {}))

    def __getattr__(self, attr):
        # TODO: only consider attr not starting with "_"?
        return getattr(self._get_default_source(), attr)

    def __repr__(self):
        return "<Catalog Entry: %s>" % self.name


import datetime
import os.path
import shutil
import tempfile
import time

import pytest

import pandas

from .util import assert_items_equal
from intake import Catalog, open_catalog
from intake.catalog import exceptions, local
from intake.catalog.local import get_dir, UserParameter


def abspath(filename):
    return os.path.join(os.path.dirname(__file__), filename)


@pytest.fixture
def catalog1():
    path = os.path.dirname(__file__)
    return Catalog(os.path.join(path, 'catalog1.yml'))


def test_local_catalog(catalog1):
    assert_items_equal(list(catalog1),
                       ['use_example1', 'nested', 'entry1', 'entry1_part',
                        'remote_env', 'local_env', 'text', 'arr'])
    assert catalog1['entry1'].describe() == {
        'container': 'dataframe',
        'direct_access': 'forbid',
        'user_parameters': [],
        'description': 'entry1 full'
    }
    assert catalog1['entry1_part'].describe() == {
        'container': 'dataframe',
        'user_parameters': [
            {
                'name': 'part',
                'description': 'part of filename',
                'default': '1',
                'type': 'str',
                'allowed': ['1', '2'],
            }
        ],
        'description': 'entry1 part',
        'direct_access': 'allow'
    }
    assert catalog1['entry1'].get().container == 'dataframe'
    md = catalog1['entry1'].get().metadata
    md.pop('catalog_dir')
    assert md == dict(foo='bar', bar=[1, 2, 3], cache=[])

    # Use default parameters
    assert catalog1['entry1_part'].get().container == 'dataframe'
    # Specify parameters
    assert catalog1['entry1_part'].get(part='2').container == 'dataframe'


def test_nested(catalog1):
    assert 'nested' in catalog1
    assert 'entry1' in catalog1.nested.nested()
    assert catalog1.entry1.read().equals(catalog1.nested.nested.entry1.read())
    assert 'nested.nested' not in catalog1.walk(depth=1)
    assert 'nested.nested' in catalog1.walk(depth=2)


def test_source_plugin_config(catalog1):
    from intake import registry
    assert 'example1' in registry
    assert 'example2' in registry


def test_metadata(catalog1):
    assert hasattr(catalog1, 'metadata')
    assert catalog1['metadata']['test'] is True


def test_use_source_plugin_from_config(catalog1):
    catalog1['use_example1'].get()


def test_get_dir():
    assert get_dir('s3://path/catalog.yml') == 's3://path'
    assert get_dir('https://example.com/catalog.yml') == 'https://example.com'
    path = 'example/catalog.yml'
    assert get_dir(path) == os.path.join(os.getcwd(), os.path.dirname(path))
    path = '/example/catalog.yml'
    assert get_dir(path) == os.path.dirname(path)
    path = 'example'
    assert get_dir(path) == os.path.join(os.getcwd(), '')


@pytest.mark.parametrize("dtype,expected", [
    ("bool", False),
    ("datetime", pandas.Timestamp(1970, 1, 1, 0, 0, 0)),
    ("float", 0.0),
    ("int", 0),
    ("list", []),
    ("str", ""),
    ("unicode", u""),
])
def test_user_parameter_default_value(dtype, expected):
    p = local.UserParameter('a', 'a desc', dtype)
    assert p.validate(None) == expected


@pytest.mark.parametrize("dtype,given,expected", [
    ("bool", "true", True),
    ("bool", 0, False),
    ("datetime", datetime.datetime(2018, 1, 1, 0, 34, 0), pandas.Timestamp(2018, 1, 1, 0, 34, 0)),
    ("datetime", "2018-01-01 12:34AM", pandas.Timestamp(2018, 1, 1, 0, 34, 0)),
    ("datetime", 1234567890000000000, pandas.Timestamp(2009, 2, 13, 23, 31, 30)),
    ("float", "3.14", 3.14),
    ("int", "1", 1),
    ("list", (3, 4), [3, 4]),
    ("str", 1, "1"),
    ("unicode", "foo", u"foo"),
])
def test_user_parameter_coerce_value(dtype, given, expected):
    p = local.UserParameter('a', 'a desc', dtype, given)
    assert p.validate(given) == expected


@pytest.mark.parametrize("given", ["now", "today"])
def test_user_parameter_coerce_special_datetime(given):
    p = local.UserParameter('a', 'a desc', 'datetime', given)
    assert type(p.validate(given)) == pandas.Timestamp


@pytest.mark.parametrize("dtype,given,expected", [
    ("float", "100.0", 100.0),
    ("int", "20", 20),
    ("int", 20.0, 20),
])
def test_user_parameter_coerce_min(dtype, given, expected):
    p = local.UserParameter('a', 'a desc', dtype, expected, min=given)
    assert p.min == expected


@pytest.mark.parametrize("dtype,given,expected", [
    ("float", "100.0", 100.0),
    ("int", "20", 20),
    ("int", 20.0, 20),
])
def test_user_parameter_coerce_max(dtype, given, expected):
    p = local.UserParameter('a', 'a desc', dtype, expected, max=given)
    assert p.max == expected


@pytest.mark.parametrize("dtype,given,expected", [
    ("float", [50, "100.0", 150.0], [50.0, 100.0, 150.0]),
    ("int", [1, "2", 3.0], [1, 2, 3]),
])
def test_user_parameter_coerce_allowed(dtype, given, expected):
    p = local.UserParameter('a', 'a desc', dtype, expected[0], allowed=given)
    assert p.allowed == expected


def test_user_parameter_validation_range():
    p = local.UserParameter('a', 'a desc', 'int', 1, min=0, max=3)

    with pytest.raises(ValueError) as except_info:
        p.validate(-1)
    assert 'less than' in str(except_info.value)

    assert p.validate(0) == 0
    assert p.validate(1) == 1
    assert p.validate(2) == 2
    assert p.validate(3) == 3

    with pytest.raises(ValueError) as except_info:
        p.validate(4)
    assert 'greater than' in str(except_info.value)


def test_user_parameter_validation_allowed():
    p = local.UserParameter('a', 'a desc', 'int', 1, allowed=[1, 2])

    with pytest.raises(ValueError) as except_info:
        p.validate(0)
    assert 'allowed' in str(except_info.value)

    assert p.validate(1) == 1
    assert p.validate(2) == 2

    with pytest.raises(ValueError) as except_info:
        p.validate(3)
    assert 'allowed' in str(except_info.value)


@pytest.mark.parametrize("filename", [
    "catalog_non_dict",
    "data_source_missing",
    "data_source_name_non_string",
    "data_source_non_dict",
    "data_source_value_non_dict",
    "params_missing_required",
    "params_name_non_string",
    "params_non_dict",
    "params_value_bad_choice",
    "params_value_bad_type",
    "params_value_non_dict",
    "plugins_non_dict",
    "plugins_source_both",
    "plugins_source_missing",
    "plugins_source_missing_key",
    "plugins_source_non_dict",
    "plugins_source_non_list",
])
def test_parser_validation_error(filename):
    with pytest.raises(exceptions.ValidationError):
        Catalog(abspath(filename + ".yml"))


@pytest.mark.parametrize("filename", [
    "obsolete_data_source_list",
    "obsolete_params_list",
])
def test_parser_obsolete_error(filename):
    with pytest.raises(exceptions.ObsoleteError):
        Catalog(abspath(filename + ".yml"))


def test_union_catalog():
    path = os.path.dirname(__file__)
    uri1 = os.path.join(path, 'catalog_union_1.yml')
    uri2 = os.path.join(path, 'catalog_union_2.yml')

    union_cat = Catalog([uri1, uri2])

    assert_items_equal(list(union_cat), ['entry1', 'entry1_part', 'use_example1'])

    assert union_cat.entry1_part.describe() == {
        'container': 'dataframe',
        'user_parameters': [
            {
                'name': 'part',
                'description': 'part of filename',
                'default': '1',
                'type': 'str',
                'allowed': ['1', '2'],
            }
        ],
        'description': 'entry1 part',
        'direct_access': 'allow'
    }

    desc_open = union_cat.entry1_part.describe_open()
    assert desc_open['args']['urlpath'].endswith('entry1_1.csv')
    del desc_open['args']['urlpath']  # Full path will be system dependent
    desc_open['args']['metadata'].pop('catalog_dir')
    assert 'csv' in str(desc_open.pop('plugin'))
    assert desc_open == {
        'args': {'metadata': {'bar': [2, 4, 6], 'cache': [], 'foo': 'baz'}},
        'description': 'entry1 part',
        'direct_access': 'allow',
        'metadata': {'bar': [2, 4, 6], 'cache': [], 'foo': 'baz'},
    }

    # Implied creation of data source
    assert union_cat.entry1.container == 'dataframe'
    md = union_cat.entry1.metadata
    md.pop('catalog_dir')
    assert md == dict(foo='bar', bar=[1, 2, 3], cache=[])

    # Use default parameters in explict creation of data source
    assert union_cat.entry1_part().container == 'dataframe'
    # Specify parameters in creation of data source
    assert union_cat.entry1_part(part='2').container == 'dataframe'


def test_empty_catalog():
    cat = Catalog()
    assert list(cat) == []


def test_nonexistent_error():
    with pytest.raises(IOError):
        local.YAMLFileCatalog('nonexistent')


def test_duplicate_data_sources():
    path = os.path.dirname(__file__)
    uri = os.path.join(path, 'catalog_dup_sources.yml')

    with pytest.raises(exceptions.DuplicateKeyError) as except_info:
        c = Catalog(uri)


def test_duplicate_parameters():
    path = os.path.dirname(__file__)
    uri = os.path.join(path, 'catalog_dup_parameters.yml')

    with pytest.raises(exceptions.DuplicateKeyError) as except_info:
        c = Catalog(uri)


@pytest.fixture
def temp_catalog_file():
    path = tempfile.mkdtemp()
    catalog_file = os.path.join(path, 'catalog.yaml')
    with open(catalog_file, 'w') as f:
        f.write('''
sources:
  a:
    driver: csv
    args:
      urlpath: /not/a/file
  b:
    driver: csv
    args:
      urlpath: /not/a/file
        ''')

    yield catalog_file

    shutil.rmtree(path)


def test_catalog_file_removal(temp_catalog_file):
    cat_dir = os.path.dirname(temp_catalog_file)
    cat = Catalog(cat_dir + '/*')
    assert set(cat) == {'a', 'b'}

    os.remove(temp_catalog_file)
    time.sleep(1.5)  # wait for catalog refresh
    assert set(cat) == set()


def test_default_expansions():
    try:
        os.environ['INTAKE_INT_TEST'] = '1'
        par = UserParameter('', '', 'int', default='env(INTAKE_INT_TEST)')
        par.expand_defaults()
        assert par.expanded_default == 1
    finally:
        del os.environ['INTAKE_INT_TEST']

    par = UserParameter('', '', 'str', default='env(USER)')
    par.expand_defaults(getenv=False)
    assert par.expanded_default == 'env(USER)'
    par.expand_defaults()
    assert par.expanded_default == os.getenv('USER', '')

    par = UserParameter('', '', 'str', default='client_env(USER)')
    par.expand_defaults()
    assert par.expanded_default == 'client_env(USER)'
    par.expand_defaults(client=True)
    assert par.expanded_default == os.getenv('USER', '')

    par = UserParameter('', '', 'str', default='shell(echo success)')
    par.expand_defaults(getshell=False)
    assert par.expanded_default == 'shell(echo success)'
    par.expand_defaults()
    assert par.expanded_default == 'success'

    par = UserParameter('', '', 'str', default='client_shell(echo success)')
    par.expand_defaults(client=True)
    assert par.expanded_default == 'success'

    par = UserParameter('', '', 'int', default=1)
    par.expand_defaults()  # no error from string ops


def test_remote_cat(http_server):
    url = http_server + 'catalog1.yml'
    cat = Catalog(url)
    assert 'entry1' in cat
    assert cat.entry1.describe()


def test_multi_plugins():
    from intake.source.csv import CSVSource
    fn = abspath('multi_plugins.yaml')
    cat = open_catalog(fn)
    s = cat.tables0()
    assert isinstance(s, CSVSource)

    s = cat.tables1()
    assert isinstance(s, CSVSource)

    s = cat.tables2()
    assert isinstance(s, CSVSource)

    s = cat.tables3()
    assert isinstance(s, CSVSource)
    assert s._csv_kwargs == {}

    s = cat.tables3(plugin='myplug')
    assert isinstance(s, CSVSource)
    assert s._csv_kwargs == {}

    s = cat.tables3(plugin='myplug2')
    assert isinstance(s, CSVSource)
    assert s._csv_kwargs is True

    with pytest.raises(ValueError):
        cat.tables4()
    with pytest.raises(ValueError):
        cat.tables4(plugin='myplug')
    with pytest.raises(ValueError):
        cat.tables4(plugin='myplug2')

    s = cat.tables5()
    assert isinstance(s, CSVSource)

    with pytest.raises(ValueError):
        cat.tables5(plugin='myplug')

    fn = abspath('multi_plugins2.yaml')
    with pytest.raises(ValueError):
        open_catalog(fn)


def test_no_plugins():
    fn = abspath('multi_plugins.yaml')
    cat = open_catalog(fn)
    s = cat.tables6
    with pytest.raises(ValueError):
        s()
    s = cat.tables7
    with pytest.raises(ValueError):
        s()



from . import base


class TextFilesSource(base.DataSource):
    """Read textfiles as sequence of lines

    Protype of sources reading sequential data.

    Takes a set of files, and returns an iterator over the text in each of them.
    The files can be local or remote. Extra parameters for encoding, etc.,
    go into ``storage_options``.
    """
    name = 'textfiles'
    version = '0.0.1'
    container = 'python'
    partition_access = True

    def __init__(self, urlpath, metadata=None,
                 storage_options=None):
        """

        Parameters
        ----------
        urlpath : str or list(str)
            Target files. Can be a glob-path (with "*") and include protocol
            specified (e.g., "s3://"). Can also be a list of absolute paths.
        storage_options: dict
            Options to pass to the file reader backend, including text-specific
            encoding arguments, and parameters specific to the remote
            file-system driver, if using.
        """
        self._urlpath = urlpath
        self._storage_options = storage_options or {}
        self._dataframe = None
        self._files = None

        super(TextFilesSource, self).__init__(metadata=metadata)

    def _get_schema(self):
        from dask.bytes import open_files
        if self._files is None:

            urlpath = self._get_cache(self._urlpath)[0]

            self._files = open_files(urlpath, mode='rt',
                                     **self._storage_options)
            self.npartitions = len(self._files)
        return base.Schema(datashape=None,
                           dtype=None,
                           shape=(None, ),
                           npartitions=self.npartitions,
                           extra_metadata=self.metadata)

    def _get_partition(self, i):
        return get_file(self._files[i])

    def read(self):
        self._get_schema()
        return self.to_dask().compute()

    def to_spark(self):
        from intake_spark.base import SparkHolder
        h = SparkHolder(False, [
            ('textFile', (self._urlpath, ))
        ], {})
        return h.setup()

    def to_dask(self):
        import dask.bag as db
        from dask import delayed
        dfile = delayed(get_file)
        return db.from_delayed([dfile(f) for f in self._files])


def get_file(f):
    """Serializable function to take an OpenFile object and read lines"""
    with f as f:
        return list(f)



import appdirs
import json
import os
import subprocess
import sys

from intake.config import conf
from .local import YAMLFilesCatalog, Catalog


def load_user_catalog():
    """Return a catalog for the platform-specific user Intake directory"""
    cat_dir = user_data_dir()
    if not os.path.isdir(cat_dir):
        return Catalog()
    else:
        return YAMLFilesCatalog(cat_dir)


def user_data_dir():
    """Return the user Intake catalog directory"""
    return appdirs.user_data_dir(appname='intake', appauthor='intake')


def load_global_catalog():
    """Return a catalog for the environment-specific Intake directory"""
    cat_dir = global_data_dir()
    if not os.path.isdir(cat_dir):
        return Catalog()
    else:
        return YAMLFilesCatalog(cat_dir)


CONDA_VAR = 'CONDA_PREFIX'
VIRTUALENV_VAR = 'VIRTUAL_ENV'


def conda_prefix():
    """Fallback: ask conda in PATH for its prefix"""
    try:
        out = subprocess.check_output(['conda', 'info', '--json'])
        return json.loads(out.decode())["default_prefix"]
    except (subprocess.CalledProcessError, json.JSONDecodeError, OSError):
        return False


def which(program):
    """Emulate posix ``which``

    https://stackoverflow.com/a/377028/3821154
    """
    import os

    def is_exe(fpath):
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

    fpath, fname = os.path.split(program)
    if fpath:
        if is_exe(program):
            return program
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return exe_file

    return None


def global_data_dir():
    """Return the global Intake catalog dir for the current environment"""
    prefix = False
    if VIRTUALENV_VAR in os.environ:
        prefix = os.environ[VIRTUALENV_VAR]
    elif CONDA_VAR in os.environ:
        prefix = sys.prefix
    elif which('conda'):
        # conda exists but is not activated
        prefix = conda_prefix()

    if prefix:
        # conda and virtualenv use Linux-style directory pattern
        return os.path.join(prefix, 'share', 'intake')
    else:
        return appdirs.site_data_dir(appname='intake', appauthor='intake')


def load_combo_catalog():
    """Load a union of the user and global catalogs for convenience"""
    user_dir = user_data_dir()
    global_dir = global_data_dir()

    cat_dirs = []
    if os.path.isdir(user_dir):
        cat_dirs.append(user_dir + '/*.yaml')
        cat_dirs.append(user_dir + '/*.yml')
    if os.path.isdir(global_dir):
        cat_dirs.append(global_dir + '/*.yaml')
        cat_dirs.append(global_dir + '/*.yml')
    for path_dir in conf.get('catalog_path', []):
        if path_dir != '':
            if not path_dir.endswith(('yaml', 'yml')):
                cat_dirs.append(path_dir + '/*.yaml')
                cat_dirs.append(path_dir + '/*.yml')
            else:
                cat_dirs.append(path_dir)

    return YAMLFilesCatalog(cat_dirs, name='builtin')



print("import: 'intake'")
import intake

print("import: 'intake.auth'")
import intake.auth

print("import: 'intake.auth.tests'")
import intake.auth.tests

print("import: 'intake.catalog'")
import intake.catalog

print("import: 'intake.catalog.tests'")
import intake.catalog.tests

print("import: 'intake.cli'")
import intake.cli

print("import: 'intake.cli.client'")
import intake.cli.client

print("import: 'intake.cli.client.tests'")
import intake.cli.client.tests

print("import: 'intake.cli.server'")
import intake.cli.server

print("import: 'intake.cli.server.tests'")
import intake.cli.server.tests

print("import: 'intake.container'")
import intake.container

print("import: 'intake.gui'")
import intake.gui

print("import: 'intake.source'")
import intake.source

print("import: 'intake.source.tests'")
import intake.source.tests

print("import: 'intake.tests'")
import intake.tests




# Base classes for Data Loader interface
from .cache import make_caches
import sys
PY2 = sys.version_info[0] == 2


class Schema(dict):
    """Holds details of data description for any type of data-source

    This should always be pickleable, so that it can be sent from a server
    to a client, and contain all information needed to recreate a RemoteSource
    on the client.
    """

    def __init__(self, **kwargs):
        """
        Parameters
        ----------
        kwargs: typically include datashape, dtype, shape
        """
        super(Schema, self).__init__(**kwargs)
        for field in ['datashape', 'dtype', 'extra_metadata', 'shape']:
            # maybe a default-dict
            if field not in self:
                self[field] = None

    def __repr__(self):
        return ("<Schema instance>\n"
                "dtype: {}\n"
                "shape: {}\n"
                "metadata: {}"
                "".format(self.dtype, self.shape, self.extra_metadata))

    def __getattr__(self, item):
        return self[item]


class DataSource(object):
    """An object which can produce data

    This is the base class for all Intake plugins, including catalogs and
    remote (server) data objects. To produce a new plugin commonly involves
    subclassing this definition and overriding some or all of the methods.

    This class is not useful in itself, most methods raise NotImplemented.
    """
    name = None
    version = None
    container = None
    partition_access = False
    datashape = None
    description = None

    @property
    def cache_dirs(self):
        return [c._cache_dir for c in self.cache]

    def set_cache_dir(self, cache_dir):
        for c in self.cache:
            c._cache_dir = cache_dir

    def __new__(cls, *args, **kwargs):
        """Capture creation args when instantiating"""
        o = object.__new__(cls)
        # automatically capture __init__ arguments for pickling
        o._captured_init_args = args
        o._captured_init_kwargs = kwargs

        # monkey for requests keywords
        for key in ['auth', 'verify']:
            if (kwargs and kwargs.get('storage_options', None)
                    and key in kwargs['storage_options']):

                if isinstance(kwargs['storage_options'][key], list):
                    kwargs['storage_options'][key] = tuple(
                        kwargs['storage_options'][key])
        return o

    def __getstate__(self):
        return dict(args=self._captured_init_args,
                    kwargs=self._captured_init_kwargs)

    def __setstate__(self, state):
        self._captured_init_kwargs = state['kwargs']
        self._captured_init_args = state['args']
        self.__init__(*state['args'], **state['kwargs'])

    def __init__(self, metadata=None):
        # default data
        self.metadata = metadata or {}
        if isinstance(self.metadata, dict):
            storage_options = self._captured_init_kwargs.get('storage_options', {})
            self.cache = make_caches(self.name, self.metadata.get('cache'),
                                     catdir=self.metadata.get('catalog_dir', None),
                                     storage_options=storage_options)
        self.datashape = None
        self.dtype = None
        self.shape = None
        self.npartitions = 0
        self._schema = None
        self.catalog_object = None
        self.on_server = False

    def _get_cache(self, urlpath):
        if len(self.cache) == 0:
            return [urlpath]
        return [c.load(urlpath) for c in self.cache]

    def _get_schema(self):
        """Subclasses should return an instance of base.Schema"""
        raise Exception('Subclass should implement _get_schema()')

    def _get_partition(self, i):
        """Subclasses should return a container object for this partition

        This function will never be called with an out-of-range value for i.
        """
        raise Exception('Subclass should implement _get_partition()')

    def _close(self):
        """Subclasses should close all open resources"""
        raise Exception('Subclass should implement _close()')

    # These methods are implemented from the above two methods and do not need
    # to be overridden unless custom behavior is required

    def _load_metadata(self):
        """load metadata only if needed"""
        if self._schema is None:
            self._schema = self._get_schema()
            self.datashape = self._schema.datashape
            self.dtype = self._schema.dtype
            self.shape = self._schema.shape
            self.npartitions = self._schema.npartitions
            self.metadata.update(self._schema.extra_metadata)

    def yaml(self, with_plugin=False):
        """Return YAML representation of this data-source

        The output may be roughly appropriate for inclusion in a YAML
        catalog. This is a best-effort implementation

        Parameters
        ----------
        with_plugin: bool
            If True, create a "plugins" section, for cases where this source
            is created with a plugin not expected to be in the global Intake
            registry.
        """
        try:
            from ruamel.yaml import dump
        except:
            from ruamel_yaml import dump
        import inspect
        kwargs = self._captured_init_kwargs.copy()
        meta = kwargs.pop('metadata', self.metadata) or {}
        if PY2:
            kwargs.update(dict(zip(inspect.getargspec(self.__init__).args,
                          self._captured_init_args)))
        else:
            kwargs.update(dict(zip(inspect.signature(self.__init__).parameters,
                                   self._captured_init_args)))
        data = {'sources': {self.name: {
            'driver': self.__class__.name,
            'description': self.description or "",
            'metadata': meta,
            'args': kwargs
        }}}
        if with_plugin:
            data['plugins'] = {
                'source': [{'module': self.__module__}]}
        return dump(data, default_flow_style=False)

    @property
    def plots(self):
        """List custom associated quick-plots """
        return list(self.metadata.get('plots', {}))

    def discover(self):
        """Open resource and populate the source attributes."""
        self._load_metadata()

        return dict(datashape=self.datashape,
                    dtype=self.dtype,
                    shape=self.shape,
                    npartitions=self.npartitions,
                    metadata=self.metadata)

    def read(self):
        """Load entire dataset into a container and return it"""
        if not self.partition_access or self.npartitions == 1:
            return self._get_partition(0)
        else:
            raise NotImplementedError

    def read_chunked(self):
        """Return iterator over container fragments of data source"""
        self._load_metadata()
        for i in range(self.npartitions):
            yield self._get_partition(i)

    def read_partition(self, i):
        """Return a (offset_tuple, container) corresponding to i-th partition.

        Offset tuple is of same length as shape.

        By default, assumes i should be an integer between zero and npartitions;
        override for more complex indexing schemes.
        """
        self._load_metadata()
        if i < 0 or i >= self.npartitions:
            raise IndexError('%d is out of range' % i)

        return self._get_partition(i)

    def to_dask(self):
        """Return a dask container for this data source"""
        raise NotImplementedError

    def to_spark(self):
        """Provide an equivalent data object in Apache Spark

        The mapping of python-oriented data containers to Spark ones will be
        imperfect, and only a small number of drivers are expected to be able
        to produce Spark objects. The standard arguments may b translated,
        unsupported or ignored, depending on the specific driver.

        This method requires the package intake-spark
        """
        raise NotImplementedError

    def close(self):
        """Close open resources corresponding to this data source."""
        self._close()

    # Boilerplate to make this object also act like a context manager
    def __enter__(self):
        self._load_metadata()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    @property
    def plot(self):
        """
        Returns a hvPlot object to provide a high-level plotting API.

        To display in a notebook, be sure to run ``intake.output_notebook()``
        first.
        """
        try:
            from hvplot import hvPlot
        except ImportError:
            raise ImportError("The intake plotting API requires hvplot."
                              "hvplot may be installed with:\n\n"
                              "`conda install -c pyviz hvplot` or "
                              "`pip install hvplot`.")
        metadata = self.metadata.get('plot', {})
        fields = self.metadata.get('fields', {})
        for attrs in fields.values():
            if 'range' in attrs:
                attrs['range'] = tuple(attrs['range'])
        metadata['fields'] = fields
        plots = self.metadata.get('plots', {})
        return hvPlot(self, custom_plots=plots, **metadata)

    @property
    def hvplot(self):
        """
        Returns a hvPlot object to provide a high-level plotting API.
        """
        return self.plot


class PatternMixin(object):
    """Helper class to provide file-name parsing abilities to a driver class"""
    @property
    def path_as_pattern(self):
        if hasattr(self, '_path_as_pattern'):
            return self._path_as_pattern
        raise KeyError('Plugin needs to set `path_as_pattern`'
                       ' before setting urlpath')

    @path_as_pattern.setter
    def path_as_pattern(self, path_as_pattern):
        self._path_as_pattern = path_as_pattern

    @property
    def urlpath(self):
        return self._urlpath

    @urlpath.setter
    def urlpath(self, urlpath):
        from .utils import path_to_glob

        if hasattr(self, '_original_urlpath'):
            self._urlpath = urlpath
            return

        self._original_urlpath = urlpath

        if self.path_as_pattern:
            self._urlpath = path_to_glob(urlpath)
        else:
            self._urlpath = urlpath

        if isinstance(self.path_as_pattern, bool):
            if isinstance(urlpath, str) and self._urlpath == urlpath:
                self.path_as_pattern = False

    @property
    def pattern(self):
        from .utils import path_to_pattern

        if isinstance(self.path_as_pattern, str):
            return self.path_as_pattern
        elif self.path_as_pattern:
            return path_to_pattern(self._original_urlpath, self.metadata)
        return


class AliasSource(DataSource):
    """Refer to another named source, unmodified

    The purpose of an Alias is to be able to refer to other source(s) in the
    same catalog, perhaps leaving the choice of which target to load up to the
    user. This source makes no sense outside of a catalog.

    In this case, the output of the target source is not modified, but this
    class acts as a prototype 'derived' source for processing the output of
    some standard driver.

    After initial discovery, the source's container and other details will be
    updated from the target; initially, the AliasSource container is not
    any standard.
    """
    container = 'other'
    version = 1
    name = 'alias'

    def __init__(self, target, mapping=None, metadata=None, **kwargs):
        """

        Parameters
        ----------
        target: str
            Name of the source to load, must be a key in the same catalog
        mapping: dict or None
            If given, use this to map the string passed as ``target`` to
            entries in the catalog
        metadata: dict or None
            Extra metadata to associate
        kwargs: passed on to the target
        """
        super(AliasSource, self).__init__(metadata)
        self.target = target
        self.mapping = mapping or {target: target}
        self.kwargs = kwargs
        self.metadata = metadata
        self.source = None

    def _get_source(self):
        if self.catalog_object is None:
            raise ValueError('AliasSource cannot be used outside a catalog')
        if self.source is None:
            self.source = self.catalog_object[self.mapping[self.target]](
                metadata=self.metadata, **self.kwargs)
            self.metadata = self.source.metadata.copy()
            self.container = self.source.container
            self.partition_access = self.source.partition_access
            self.description = self.source.description
            self.datashape = self.source.datashape

    def discover(self):
        self._get_source()
        return self.source.discover()

    def read(self):
        self._get_source()
        return self.source.read()

    def read_partition(self, i):
        self._get_source()
        return self.source.read_partition(i)

    def read_chunked(self):
        self._get_source()
        return self.source.read_chunked()

    def to_dask(self):
        self._get_source()
        return self.source.to_dask()



try:
    import ipywidgets
    from .widgets import DataBrowser

except ImportError:

    class DataBrowser(object):
        def __repr__(self):
            raise RuntimeError("Please install ipywidgets to use the Data "
                               "Browser")



from collections import OrderedDict
from contextlib import contextmanager
import ipywidgets as widgets
import os
import intake


here = os.path.abspath(os.path.dirname(__file__))
logo_file = os.path.join(here, 'logo.png')
logo = widgets.Box([widgets.Image.from_file(logo_file)],
                   layout=widgets.Layout(width='46px', height='35px',
                                         overflow_y='visible',
                                         overflow_x='visible'))


@contextmanager
def ignore(ob):
    try:
        ob.ignore = True
        yield
    finally:
        ob.ignore = False


class Base(object):
    done_callback = None

    def stop(self, ok=False):
        done = self.done_callback is not None
        if ok and done:
            self.done_callback(ok)
        elif done:
            self.done_callback(None)

    def __repr__(self):
        return ("Intake GUI instance: to get widget to display, you must "
                "install ipy/jupyter-widgets, run in a notebook and, in "
                "the case of jupyter-lab, install the jlab extension.")

    def _ipython_display_(self, **kwargs):
        from IPython.core.interactiveshell import InteractiveShell
        import IPython

        # from IPython.Widget._ipython_display_
        if InteractiveShell.initialized():
            if self.widget._view_name is not None:

                plaintext = repr(self)
                data = {
                    'text/plain': plaintext,
                    'application/vnd.jupyter.widget-view+json': {
                        'version_major': 2,
                        'version_minor': 0,
                        'model_id': self.widget._model_id
                    }
                }
                IPython.display.display(data, raw=True)
                self.widget._handle_displayed(**kwargs)


class DataBrowser(Base):
    """Intake data set browser

    Examples
    --------

    Display GUI in a notebook thus:

    >>> gui = intake.gui.widgets.DataBrowser()
    >>> gui.widget

    Optionally you can specify some Catalog instances to begin with. If none
    are given, will use the default one, ``intake.cat``.

    After interacting, the value of ``gui.item`` will be the entry selected.
    """

    def __init__(self, cats=None):
        if isinstance(cats, list):
            cats = cats
        elif cats is None:
            cats = [intake.cat]
        else:
            cats = [cats]
        self.cats = OrderedDict([(cat.name, cat) for cat in cats])
        self.cat_list = widgets.Select(rows=9)
        self.item = None
        self.ignore = False
        self.exception = None
        self.item_list = widgets.Select(rows=9)
        self.detail = widgets.Textarea(rows=7,
                                       placeholder='Item Description')
        self.add = widgets.Button(
            icon='plus-square',
            tooltip='Add Catalog',
            layout=widgets.Layout(width='auto'))
        self.add.on_click(self.open)
        self.search = widgets.Button(
            icon='search',
            tooltip='Search entries',
            layout=widgets.Layout(width='auto', height='auto',
                                  overflow_y='visible', overflow_x='visible'))
        self.search.on_click(self.opensearch)
        self.remove = widgets.Button(
            icon='minus-square',
            tooltip='Remove cat',
            layout=widgets.Layout(width='auto', height='auto',
                                  overflow_y='visible', overflow_x='visible'))
        self.remove.on_click(self.del_cat)
        self.logobox = widgets.VBox(
            children=[logo, self.search, self.add, self.remove],
            layout=widgets.Layout(width='auto',
                                  height='auto',
                                  overflow_y='visible',
                                  overflow_x='visible'))
        self.mid = widgets.HBox(children=[self.logobox, self.cat_list,
                                          self.item_list, self.detail])
        self.widget = widgets.VBox(children=[self.mid])
        self.widget.__repr__ = self.__repr__
        self.update_cat_list()
        self.cat_selected({'new': list(self.cats)[0]})
        self.cat_list.observe(self.cat_selected, 'value')
        self.item_list.observe(self.item_selected, 'value')

    def update_cat_list(self):
        with ignore(self):
            # only list cats with entries
            self.cat_list.options = [c for c in list(self.cats)
                                     if list(self.cats[c])]
            if len(self.cat_list.options) == 0:
                self.item_list.options = ()
                self.detail.value = ""

    def cat_selected(self, ev):
        name = ev['new']
        if self.ignore or name is None:
            return
        with ignore(self):
            names = [n + '  ->' if self.cats[name][n]._container == 'catalog'
                     else n for n in self.cats[name]]
            self.item_list.options = names
            if names:
                # avoid error if cat contains no entries
                self.item_selected({'new': names[0]}, first=True)

    def item_selected(self, ev, first=False):
        name = ev['new']
        if (self.ignore and not first) or name is None:
            return
        with ignore(self):
            if name.endswith('  ->'):
                if not first:
                    cat = self.cats[self.cat_list.value][name[:-4]]
                    self.cats[self.cat_list.value + '.' + cat.name] = cat()
                    self.update_cat_list()
            else:
                self.item = self.cats[self.cat_list.value][name]
                self.detail.value = "\n".join(
                    ["{}: {}".format(k, v) for k, v
                     in tuple(self.item.describe().items()) +
                     tuple(self.item.describe_open().items())])

    def add_cat(self, ev=None):
        """Add catalog instance to the browser

        Also called, without a catalog instance, when the (+) button is
        pressed, which will attempt to read a catalog and, if successful,
        add it to the browser.
        """
        if isinstance(ev, intake.catalog.Catalog):
            cat = ev
        else:
            try:
                cat = intake.open_catalog(ev)
                list(cat)  # fails if parse is bad
            except Exception as e:
                self.exception = e
                return
        self.cats[cat.name] = cat
        self.update_cat_list()
        self.cat_list.value = cat._name
        self.cat_selected({'new': cat._name})

    def del_cat(self, ev):
        if self.cat_list.value:
            self.cats.pop(self.cat_list.value)
        self.update_cat_list()
        if self.cat_list.options:
            cat = self.cat_list.options[-1]
            self.cat_list.value = cat
            self.cat_selected({'new': cat})

    def open(self, ev):
        cat_open = AddCat(self.file_chosen, filters=['.yaml', '.yml'])
        self.widget.children = [self.mid, cat_open.widget]

    def opensearch(self, ev):
        if self.cats:
            s = Search(done_callback=self.dosearch)
            self.widget.children = [self.mid, s.widget]

    def dosearch(self, out):
        if out:
            text, depth = out
            if len(depth) == 1:
                depth = int(depth)
            else:
                depth = 99
            cat = self.cats[self.cat_list.value]
            cat2 = cat.search(text, depth=depth)
            if len(list(cat2)):
                self.add_cat(cat2)
            self.widget.children = [self.mid]

    def file_chosen(self, fn):
        """Callback from AddCat panel"""
        if fn is not None:
            self.add_cat(fn)
        self.widget.children = [self.mid]


class Search(Base):
    def __init__(self, done_callback=None):
        self.done_callback = done_callback
        self.label = widgets.Label(value='Search Text:')
        self.search = widgets.Text(
            placeholder="Set of words",
            layout=widgets.Layout(flex='10 1 auto', width='auto'))
        self.depth = widgets.Dropdown(description="Depth",
                                      options=['1', '2', '3', '4', '5', "All"])
        self.x = widgets.Button(
            icon='close', tooltip='Close Selector',
            layout=widgets.Layout(flex='1 1 auto', width='auto'))
        self.x.on_click(lambda ev: self.stop())
        self.ok = widgets.Button(
            icon='check', tooltip='OK',
            layout=widgets.Layout(flex='1 1 auto', width='auto'))
        self.ok.on_click(lambda ev: self.stop(ok=(self.search.value,
                                                  self.depth.value)))
        self.widget = widgets.HBox(children=[self.label, self.search,
                                             self.depth, self.ok,
                                             self.x])


class AddCat(Base):
    """Sub-widget for adding new cats from file or remote"""

    def __init__(self, done_callback=None, filters=None):
        """
        Parameters
        ----------
        done_callback: function
            Called when the tick or cross buttons are clicked. Expects
            signature func(path, ok=True|False).
        """
        self.fs = FileSelector(done_callback=done_callback,
                               filters=filters or ['yaml', 'yml'])
        self.url = URLSelector(done_callback=done_callback)
        self.widget = widgets.Tab(children=[self.fs.widget, self.url.widget],)
        self.widget.set_title(0, "Local")
        self.widget.set_title(1, "Remote")


class URLSelector(Base):
    def __init__(self, done_callback=None):
        self.done_callback = done_callback
        self.lurl = widgets.Label(value='URL:')
        self.url = widgets.Text(
            placeholder="Full URL with protocol",
            layout=widgets.Layout(flex='10 1 auto', width='auto'))
        self.x = widgets.Button(
            icon='close', tooltip='Close Selector',
            layout=widgets.Layout(flex='1 1 auto', width='auto'))
        self.x.on_click(lambda ev: self.stop())
        self.ok = widgets.Button(
            icon='check', tooltip='OK',
            layout=widgets.Layout(flex='1 1 auto', width='auto'))
        self.ok.on_click(lambda ev: self.stop(ok=self.url.value))
        self.widget = widgets.HBox(children=[self.lurl, self.url, self.ok,
                                             self.x])


class FileSelector(Base):
    """
    ipywidgets interface for picking files

    The current path is stored in .path anf the current selection is stored in
    .label2.value.
    """
    def __init__(self, done_callback=None, filters=None):
        """

        Parameters
        ----------
        done_callback: function
            Called when the tick or cross buttons are clicked. Expects
            signature func(path, ok=True|False).
        filters: list of str or None
            Only show files ending in one of these strings. Normally used for
            picking file extensions. None is an alias for [''], passes all
            files.
        """
        self.done_callback = done_callback
        self.filters = [''] if filters is None else filters
        self.path = os.getcwd() + '/'
        self.main = widgets.Select(rows=15)
        self.button = widgets.Button(
            icon='chevron-left', tooltip='Parent',
            layout=widgets.Layout(flex='1 1 auto', width='auto'))
        self.button.on_click(self.up)
        self.label = widgets.Label(
            layout=widgets.Layout(flex='100 1 auto', width='auto'))
        self.x = widgets.Button(
            icon='close', tooltip='Close Selector',
            layout=widgets.Layout(width='auto'))
        self.x.on_click(lambda ev: self.stop())
        self.ok = widgets.Button(
            icon='check', tooltip='OK',
            layout=widgets.Layout(width='auto'))
        self.ok.on_click(lambda ev: self._ok())
        self.make_options()
        self.main.observe(self.changed, 'value')
        self.upper = widgets.Box(children=[self.button, self.label])
        self.right = widgets.VBox(children=[self.x, self.ok])
        self.lower = widgets.HBox(children=[self.main, self.right])
        self.widget = widgets.VBox(children=[self.upper, self.lower])
        self.ignore = False

    def _ok(self):
        fn = self.main.value
        if fn.endswith('/'):
            self.stop()
        else:
            self.stop(os.path.join(self.path, fn))

    def make_options(self):
        self.ignore = True
        self.label.value = self.path
        out = []
        for f in sorted(os.listdir(self.path)):
            if os.path.isdir(self.path + f):
                out.append(f + '/')
            elif (not f.startswith('.') and any(f.endswith(ext)
                                                for ext in self.filters)):
                out.append(f)
        self.main.value = None
        self.main.options = out
        self.ignore = False

    def up(self, ev):
        self.path = os.path.dirname(
            self.path.rstrip('/')).rstrip('/') + '/'
        self.make_options()

    def changed(self, ev):
        if self.ignore:
            return
        with ignore(self):
            fn = ev['new']
            if fn.endswith('/'):
                self.path = self.path + fn
                self.make_options()



import os
import pytest
import requests

import intake
from intake import config
from intake.util_tests import temp_conf, server


@pytest.mark.parametrize('conf', [
    {},
    {'port': 5000},
    {'other': True}
])
def test_load_conf(conf):
    inconf = config.conf.copy()
    expected = inconf.copy()
    with temp_conf(conf) as fn:
        config.load_conf(fn)
        expected.update(conf)
        assert config.conf == expected
        config.reset_conf()
        assert config.conf == inconf


# Tests with a real separate server process


def test_basic():
    with temp_conf({}) as fn:
        env = os.environ.copy()
        env['INTAKE_CONF_FILE'] = fn
        with server(env=env, wait=5000):
            r = requests.get('http://localhost:5000/v1/info')
            assert r.ok


def test_cli():
    with temp_conf({}) as fn:
        env = os.environ.copy()
        env['INTAKE_CONF_FILE'] = fn
        with server(args=('-p', '5555'), env=env, wait=5555):
            r = requests.get('http://localhost:5555/v1/info')
            assert r.ok


def test_conf():
    with temp_conf({'port': 5555}) as fn:
        env = os.environ.copy()
        env['INTAKE_CONF_FILE'] = fn
        with server(env=env, wait=5555):
            r = requests.get('http://localhost:5555/v1/info')
            assert r.ok


def test_conf_auth():
    with temp_conf({'auth': {'class': 'intake.auth.secret.SecretAuth',
                             'kwargs': {'secret': 'test'}}}) as fn:
        env = os.environ.copy()
        env['INTAKE_CONF_FILE'] = fn
        with server(env=env, wait=5000):
            # raw request
            r = requests.get('http://localhost:5000/v1/info')
            assert r.status_code == 403
            r = requests.get('http://localhost:5000/v1/info',
                             headers={'intake-secret': 'test'})
            assert r.ok

            # with cat
            with pytest.raises(Exception):
                intake.Catalog('intake://localhost:5000')

            cat = intake.Catalog('intake://localhost:5000',
                                 storage_options={'headers':
                                                  {'intake-secret': 'test'}})
            assert 'entry1' in cat



from __future__ import print_function

import argparse
import logging
import os
import signal
import sys

import tornado.ioloop
import tornado.web

from .server import IntakeServer
logger = logging.getLogger('intake')


def call_exit_on_sigterm(signal, frame):
    sys.exit(0)


def main(argv=None):
    from intake.config import conf
    from intake import Catalog

    if argv is None:
        argv = sys.argv

    parser = argparse.ArgumentParser(description='Intake Catalog Server')
    parser.add_argument('-p', '--port', type=int, default=conf['port'],
                        help='port number for server to listen on')
    parser.add_argument('--sys-exit-on-sigterm', action='store_true',
                        help='internal flag used during unit testing to ensure .coverage file is written')
    parser.add_argument('catalog_args', metavar='FILE', type=str, nargs='+',
                        help='Name of catalog YAML file')
    args = parser.parse_args(argv[1:])

    if args.sys_exit_on_sigterm:
        signal.signal(signal.SIGTERM, call_exit_on_sigterm)

    logger.info('Creating catalog from:')
    for arg in args.catalog_args:
        logger.info('  - %s' % arg)

    catargs = args.catalog_args
    catargs = catargs[0] if len(catargs) == 1 else catargs
    logger.info("catalog_args: %s" % catargs)
    catalog = Catalog(catargs)

    logger.info('Entries:' + ','.join(list(catalog)))

    logger.info('Listening on port %d' % args.port)

    server = IntakeServer(catalog)
    app = server.make_app()
    server.start_periodic_functions(close_idle_after=3600.0)

    app.listen(args.port)
    tornado.ioloop.IOLoop.current().start()


if __name__ == "__main__":
    sys.exit(main(sys.argv))



import itertools
from .base import RemoteSource, Schema, get_partition


class RemoteArray(RemoteSource):
    """nd-array on an Intake server"""
    name = 'remote_ndarray'
    container = 'ndarray'

    def __init__(self, url, headers, **kwargs):
        super(RemoteArray, self).__init__(url, headers, **kwargs)
        self.npartitions = kwargs['npartitions']
        self.shape = tuple(kwargs['shape'])
        self.metadata = kwargs['metadata']
        self.dtype = kwargs['dtype']
        self.chunks = tuple(tuple(c)
                            for c in tuple(kwargs.get('chunks', (-1, ))))
        self.arr = None
        self._schema = Schema(npartitions=self.npartitions,
                              extra_metadata=self.metadata,
                              dtype=self.dtype,
                              shape=self.shape,
                              chunks=self.chunks,
                              datashape=None)

    def _load_metadata(self):
        import dask.array as da
        if self.arr is None:
            name = 'remote-array-' + self._source_id
            nparts = (range(len(n)) for n in self.chunks)
            dask = {
                (name, ) + part: (get_partition, self.url, self.headers,
                                  self._source_id, self.container, part)

                for part in itertools.product(*nparts)
            }
            self.arr = da.Array(dask, name, self.chunks, self.dtype, self.shape)

        return self._schema

    def _get_partition(self, i):
        self._load_metadata()
        return self.arr.blocks[i].compute()

    def read_partition(self, i):
        self._load_metadata()
        return self._get_partition(i)

    def read(self):
        self._load_metadata()
        return self.arr.compute()

    def to_dask(self):
        self._load_metadata()
        return self.arr

    def _close(self):
        self.arr = None



from __future__ import absolute_import
import re
import warnings

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions
from . import source
from .source.base import Schema, DataSource
from .catalog.base import Catalog, RemoteCatalog
from .catalog import local
from .catalog.default import load_combo_catalog
from .source import registry
from .source.discovery import autodiscover
from .gui import DataBrowser

# Populate list of autodetected plugins
registry.update(autodiscover())

from .source import csv, textfiles, npy
registry['csv'] = csv.CSVSource
registry['textfiles'] = textfiles.TextFilesSource
registry['catalog'] = Catalog
registry['intake_remote'] = RemoteCatalog
registry['numpy'] = npy.NPySource


def make_open_functions():
    """From the current state of ``registry``, create open_* functions"""
    # Create shortcut open methods
    if hasattr(str, 'isidentifier'):
        def isidentifier(x):
            return x.isidentifier()
    else:
        IDENTIFIER_REGEX = re.compile(r'[a-zA-Z_][a-zA-Z0-9_]*$')
        isidentifier = IDENTIFIER_REGEX.match
    for plugin_name, plugin in registry.items():
        func_name = 'open_' + plugin_name
        if not isidentifier(func_name):
            # primitive name normalization
            func_name = re.sub('[-=~^&|@+]', '_', func_name)
        if isidentifier(func_name):
            globals()[func_name] = plugin
        else:
            warnings.warn('Invalid Intake plugin name "%s" found.' %
                          plugin_name)


def output_notebook(inline=True, logo=False):
    """
    Load the notebook extension

    Parameters
    ----------
    inline : boolean (optional)
        Whether to inline JS code or load it from a CDN
    logo : boolean (optional)
        Whether to show the logo(s)
    """
    try:
        import hvplot
    except ImportError:
        raise ImportError("The intake plotting API requires hvplot."
                          "hvplot may be installed with:\n\n"
                          "`conda install -c pyviz hvplot` or "
                          "`pip install hvplot`.")
    import holoviews as hv
    return hv.extension('bokeh', inline=inline, logo=logo)


make_open_functions()
cat = load_combo_catalog()


def open_catalog(uri=None, **kwargs):
    """Create a Catalog object

    Can load YAML catalog files, connect to an intake server, or create any
    arbitrary Catalog subclass instance. In the general case, the user should
    supply ``driver=`` with a value from the plugins registry which has a
    container type of catalog. File locations can generally be remote, if
    specifying a URL protocol.

    The default behaviour if not specifying the driver is as follows:

    - if ``uri`` is a a single string ending in "yml" or "yaml", open it as a
      catalog file
    - if ``uri`` is a list of strings, a string containing a glob character
      ("*") or a string not ending in "y(a)ml", open as a set of catalog
      files. In the latter case, assume it is a directory.
    - if ``uri`` beings with protocol ``"intake:"``, connect to a remote
      Intake server
    - otherwise, create a base Catalog object without entries.

    Parameters
    ----------
    uri: str
        Designator for the location of the catalog.
    kwargs:
        passed to subclass instance, see documentation of the individual
        catalog classes.
    """
    driver = kwargs.pop('driver', None)
    if driver is None:
        if uri:
            if ((isinstance(uri, str) and "*" in uri)
                    or ((isinstance(uri, (list, tuple))) and len(uri) > 1)):
                # glob string or list of files/globs
                driver = 'yaml_files_cat'
            elif isinstance(uri, (list, tuple)) and len(uri) == 1:
                uri = uri[0]
                if "*" in uri[0]:
                    # single glob string in a list
                    driver = 'yaml_files_cat'
                else:
                    # single filename in a list
                    driver = 'yaml_file_cat'
            elif isinstance(uri, str):
                # single URL
                if uri.startswith('intake:'):
                    # server
                    driver = 'intake_remote'
                else:
                    if uri.endswith(('.yml', '.yaml')):
                        driver = 'yaml_file_cat'
                    else:
                        uri = uri + '/*'
                        driver = 'yaml_files_cat'
        else:
            # empty cat
            driver = 'catalog'
    if driver not in registry:
        raise ValueError('Unknown catalog driver, supply one of: %s'
                         % list(registry))
    return registry[driver](uri, **kwargs)


Catalog = open_catalog
gui = DataBrowser()



import intake
import intake.config
from intake.source.cache import CacheMetadata
import os
import pytest
import subprocess
import sys
from intake.source.tests.util import temp_cache
cpath = os.path.abspath(
    os.path.join(os.path.dirname(__file__),
                 '../../../catalog/tests/catalog_caching.yml'))


@pytest.mark.skipif(sys.version_info[0] == 2,
                    reason="Py2 exists early on argparse")
def test_help(tempdir):
    out = subprocess.check_output('INTAKE_CONF_DIR=%s intake cache' % tempdir,
                                  shell=True).decode()
    assert 'usage: ' in out

    out2 = subprocess.check_output('INTAKE_CONF_DIR=%s intake cache -h' % tempdir,
                                   shell=True).decode()
    assert out2 == out


def test_list_keys(temp_cache):
    tempdir = intake.config.confdir
    out = subprocess.check_output('INTAKE_CONF_DIR=%s intake cache'
                                  ' list-keys' % tempdir,
                                  shell=True).decode()
    assert '[]\n' in out  # empty cache
    cat = intake.open_catalog(cpath)
    cat.test_cache.read()
    out = subprocess.check_output('INTAKE_CONF_DIR=%s intake cache list-keys'
                                  '' % tempdir,
                                  shell=True).decode()
    assert 'states.csv' in out


def test_precache(temp_cache):
    tempdir = intake.config.confdir
    out = subprocess.check_output('INTAKE_CONF_DIR=%s intake cache list-keys'
                                  '' % tempdir,
                                  shell=True).decode()
    assert out == "[]\n\n"
    out = subprocess.check_output('INTAKE_CONF_DIR=%s INTAKE_CACHE_DIR=%s '
                                  'intake precache %s ' %
                                  (tempdir, tempdir, cpath), shell=True).decode()
    assert out.count('Caching for entry') > 1
    out = subprocess.check_output('INTAKE_CONF_DIR=%s intake cache list-keys'
                                  '' % tempdir,
                                  shell=True).decode()
    assert 'states.csv' in out
    assert 'small.npy' in out


def test_clear_all(temp_cache):
    tempdir = intake.config.confdir
    cat = intake.open_catalog(cpath)
    cat.test_cache.read()
    md = CacheMetadata()
    assert len(md) == 1
    assert 'states' in list(md)[0]
    subprocess.call('INTAKE_CONF_DIR=%s intake cache clear'
                    '' % tempdir,
                    shell=True)
    md = CacheMetadata()
    assert len(md) == 0


def test_clear_one(temp_cache):
    tempdir = intake.config.confdir
    cat = intake.open_catalog(cpath)
    cat.test_cache.read()
    cat.arr_cache.read()
    md = CacheMetadata()
    keys = list(md)
    assert len(keys) == 2
    subprocess.call('INTAKE_CONF_DIR=%s intake cache clear %s'
                    '' % (tempdir, keys[0]),
                    shell=True)
    md = CacheMetadata()
    assert len(md) == 1
    assert list(md)[0] == keys[1]


def test_usage(temp_cache):
    tempdir = intake.config.confdir
    from intake.source.cache import BaseCache
    BaseCache(None, None).clear_all()
    out = subprocess.check_output('INTAKE_CONF_DIR=%s INTAKE_CACHE_DIR=%s'
                                  ' intake cache usage' % (tempdir, tempdir),
                                  shell=True).decode()
    assert '0.0' in out  # empty!



import os.path
import pickle

import pytest
import pandas as pd

from ...source.tests.util import verify_datasource_interface
from .util import assert_items_equal
from intake import Catalog

TEST_CATALOG_PATH = os.path.join(os.path.dirname(__file__), 'catalog1.yml')


def test_info_describe(intake_server):
    catalog = Catalog(intake_server)

    assert_items_equal(list(catalog), ['use_example1', 'nested', 'entry1',
                                       'entry1_part', 'remote_env',
                                       'local_env', 'text', 'arr'])

    info = catalog['entry1'].describe()

    assert info == {
        'container': 'dataframe',
        'description': 'entry1 full',
        'name': 'entry1',
        'direct_access': 'forbid',
        'user_parameters': []
    }

    info = catalog['entry1_part'].describe()

    assert info['direct_access'] == 'allow'


def test_bad_url(intake_server):
    bad_url = intake_server + '/nonsense_prefix'

    with pytest.raises(Exception):
        Catalog(bad_url)


def test_metadata(intake_server):
    catalog = Catalog(intake_server)
    assert hasattr(catalog, 'metadata')
    assert catalog['metadata']['test'] is True
    assert catalog.version == 1


def test_unknown_source(intake_server):
    catalog = Catalog(intake_server)

    with pytest.raises(Exception):
        catalog['does_not_exist'].describe()


def test_remote_datasource_interface(intake_server):
    catalog = Catalog(intake_server)

    d = catalog['entry1'].get()

    verify_datasource_interface(d)


def test_environment_evaluation(intake_server):
    catalog = Catalog(intake_server)
    import os
    os.environ['INTAKE_TEST'] = 'client'
    d = catalog['remote_env']


def test_read(intake_server):
    catalog = Catalog(intake_server)

    d = catalog['entry1'].get()

    test_dir = os.path.dirname(__file__)
    file1 = os.path.join(test_dir, 'entry1_1.csv')
    file2 = os.path.join(test_dir, 'entry1_2.csv')
    expected_df = pd.concat((pd.read_csv(file1), pd.read_csv(file2)))
    meta = expected_df[:0]

    info = d.discover()
    assert info['datashape'] is None
    assert info['dtype'] == {k: str(v) for k, v
                             in meta.dtypes.to_dict().items()}
    assert info['npartitions'] == 2
    assert info['shape'] == (None, 3)  # Do not know CSV size ahead of time

    md = d.metadata.copy()
    md.pop('catalog_dir', None)
    assert md == dict(foo='bar', bar=[1, 2, 3], cache=[])

    df = d.read()

    assert expected_df.equals(df)


def test_read_direct(intake_server):
    catalog = Catalog(intake_server)

    d = catalog['entry1_part'].get(part='2')
    test_dir = os.path.dirname(__file__)
    file2 = os.path.join(test_dir, 'entry1_2.csv')
    expected_df = pd.read_csv(file2)
    meta = expected_df[:0]

    info = d.discover()

    assert info['datashape'] is None
    assert info['dtype'] == {k: str(v) for k, v
                             in meta.dtypes.to_dict().items()}
    assert info['npartitions'] == 1
    assert info['shape'] == (None, 3)  # Do not know CSV size ahead of time
    md = info['metadata'].copy()
    md.pop('catalog_dir', None)
    assert md == {'bar': [2, 4, 6], 'foo': 'baz', 'cache': []}

    md = d.metadata.copy()
    md.pop('catalog_dir', None)
    assert md == dict(foo='baz', bar=[2, 4, 6], cache=[])
    assert d.description == 'entry1 part'
    df = d.read()

    assert expected_df.equals(df)


def test_read_chunks(intake_server):
    catalog = Catalog(intake_server)

    d = catalog.entry1.get()

    chunks = list(d.read_chunked())
    assert len(chunks) == 2

    test_dir = os.path.dirname(__file__)
    file1 = os.path.join(test_dir, 'entry1_1.csv')
    file2 = os.path.join(test_dir, 'entry1_2.csv')
    expected_df = pd.concat((pd.read_csv(file1), pd.read_csv(file2)))

    assert expected_df.equals(pd.concat(chunks))


def test_read_partition(intake_server):
    catalog = Catalog(intake_server)

    d = catalog.entry1.get()

    p2 = d.read_partition(1)
    p1 = d.read_partition(0)

    test_dir = os.path.dirname(__file__)
    file1 = os.path.join(test_dir, 'entry1_1.csv')
    file2 = os.path.join(test_dir, 'entry1_2.csv')
    assert pd.read_csv(file1).equals(p1)
    assert pd.read_csv(file2).equals(p2)


def test_close(intake_server):
    catalog = Catalog(intake_server)

    d = catalog.entry1.get()
    d.close()


def test_with(intake_server):
    catalog = Catalog(intake_server)

    with catalog.entry1.get() as f:
        assert f.discover()


def test_pickle(intake_server):
    catalog = Catalog(intake_server)

    d = catalog.entry1.get()

    new_d = pickle.loads(pickle.dumps(d, pickle.HIGHEST_PROTOCOL))

    df = new_d.read()

    test_dir = os.path.dirname(__file__)
    file1 = os.path.join(test_dir, 'entry1_1.csv')
    file2 = os.path.join(test_dir, 'entry1_2.csv')
    expected_df = pd.concat((pd.read_csv(file1), pd.read_csv(file2)))

    assert expected_df.equals(df)


def test_to_dask(intake_server):
    catalog = Catalog(intake_server)
    d = catalog.entry1.get()
    df = d.to_dask()

    assert df.npartitions == 2


def test_remote_env(intake_server):
    import os
    os.environ['INTAKE_TEST'] = 'client'
    catalog = Catalog(intake_server)
    with pytest.raises(Exception) as e:
        catalog.remote_env.get()
    assert 'path-server' in str(e.value)

    with pytest.raises(Exception) as e:
        catalog.local_env.get()
    assert 'path-client' in str(e.value)

    # prevents *client* from getting env
    catalog = Catalog(intake_server, getenv=False)
    with pytest.raises(Exception) as e:
        catalog.local_env.get()
    assert 'INTAKE_TEST' in str(e.value)


def test_remote_sequence(intake_server):
    import glob
    d = os.path.dirname(TEST_CATALOG_PATH)
    catalog = Catalog(intake_server)
    assert 'text' in catalog
    s = catalog.text()
    s.discover()
    assert s.npartitions == len(glob.glob(os.path.join(d, '*.yml')))
    assert s.read_partition(0)
    assert s.read()


def test_remote_arr(intake_server):
    catalog = Catalog(intake_server)
    assert 'arr' in catalog
    s = catalog.arr()
    s.discover()
    assert 'remote-array' in s.to_dask().name
    assert s.npartitions == 2
    assert s.read_partition(0).ravel().tolist() == list(range(50))
    assert s.read().ravel().tolist() == list(range(100))



import logging
logger = logging.getLogger('intake')

# The registry is the mapping of plugin name (aka driver) to DataSource class
registry = {}

# A set of fully-qualifies package.module.Class mappings
classes = {}


def get_plugin_class(name):
    if name in registry:
        return registry[name]
    if '.' not in name:
        logger.debug('Plugin name "%s" not known' % name)
        return None
    if name not in classes:
        import importlib
        mod, cls = name.rsplit('.', 1)
        try:
            module = importlib.import_module(mod)
        except ImportError:
            logger.debug('Import of plugin module "%s" failed' % mod)
            return None
        try:
            classes[name] = getattr(module, cls)
        except (KeyError, NameError):
            logger.debug('Class "%s" not found in module "%s"' % (cls, mod))
    return classes.get(name, None)



from .dataframe import RemoteDataFrame
from .ndarray import RemoteArray
from .semistructured import RemoteSequenceSource
from ..catalog.base import RemoteCatalog

# each container type is represented in the remote by one of the classes in
# this dictionary
container_map = {
    'dataframe': RemoteDataFrame,
    'python': RemoteSequenceSource,
    'ndarray': RemoteArray,
    'catalog': RemoteCatalog
}

__all__ = ['container_map']



from collections import OrderedDict
import gzip
import io
import pickle

import snappy
import msgpack


class NoneCompressor(object):
    name = 'none'

    def compress(self, data):
        return data

    def decompress(self, data):
        return data


class GzipCompressor(object):
    name = 'gzip'

    def compress(self, data):
        buf = io.BytesIO()
        with gzip.GzipFile(fileobj=buf, mode='wb', compresslevel=1) as f:
            f.write(data)
        return buf.getvalue()

    def decompress(self, data):
        with gzip.GzipFile(fileobj=io.BytesIO(data)) as f:
            return f.read()


class SnappyCompressor(object):
    name = 'snappy'

    def compress(self, data):
        return snappy.compress(data)

    def decompress(self, data):
        return snappy.decompress(data)


class MsgPackSerializer(object):
    # TODO: This is ugly, should maybe transition to
    #  distributed.protocol.serialize
    name = 'msgpack'

    def encode(self, obj, container):
        if container in ['ndarray', 'xarray']:
            import msgpack_numpy
            return msgpack.packb(obj, default=msgpack_numpy.encode)
        elif container == 'dataframe':
            import pandas as pd
            return obj.to_msgpack()
        else:
            return msgpack.packb(obj, use_bin_type=True)

    def decode(self, bytestr, container):
        from ..compat import unpack_kwargs
        if container in ['ndarray', 'xarray']:
            import msgpack_numpy
            return msgpack.unpackb(bytestr, object_hook=msgpack_numpy.decode)
        elif container == 'dataframe':
            import pandas as pd
            return pd.read_msgpack(bytestr)
        else:
            return msgpack.unpackb(bytestr, **unpack_kwargs)


class PickleSerializer(object):
    def __init__(self, protocol_level):
        self._protocol_level = protocol_level
        self.name = 'pickle%d' % protocol_level

    def encode(self, obj, container):
        return pickle.dumps(obj, protocol=self._protocol_level)

    def decode(self, bytestr, container):
        return pickle.loads(bytestr)


class ComboSerializer(object):
    def __init__(self, format_encoder, compressor):
        self._format_encoder = format_encoder
        self._compressor = compressor
        self.format_name = format_encoder.name
        self.compressor_name = compressor.name

    def encode(self, obj, container):
        return self._compressor.compress(
            self._format_encoder.encode(obj, container))

    def decode(self, bytestr, container):
        return self._format_encoder.decode(
            self._compressor.decompress(bytestr), container)


# Insert in preference order
picklers = [PickleSerializer(protocol) for protocol in [2, 1]]
serializers = [MsgPackSerializer()] + picklers
format_registry = OrderedDict([(e.name, e) for e in serializers])

compressors = [SnappyCompressor(), GzipCompressor(), NoneCompressor()]
compression_registry = OrderedDict([(e.name, e) for e in compressors])



from intake.source.base import DataSource


class ExampleSource(DataSource):
    name = 'example1'
    version = '0.1'
    container = 'dataframe'
    partition_access = True

    def __init__(self, **kwargs):
        super(ExampleSource, self).__init__()



from datetime import datetime
from hashlib import md5

import collections
import json
import logging
import os
import shutil
import warnings

from dask.bytes.utils import infer_storage_options
from intake.config import conf

logger = logging.getLogger('intake')


def sanitize_path(path):
    "Utility for cleaning up paths."

    storage_option = infer_storage_options(path)

    protocol = storage_option['protocol']
    if protocol in ('http', 'https'):
        # Most FSs remove the protocol but not HTTPFS. We need to strip
        # it to match properly.
        return os.path.normpath(path.replace("{}://".format(protocol), ''))
    elif protocol == 'file':
        # Just removing trailing slashes from file paths.
        return os.path.normpath(path)
    # Otherwise we leave the path alone
    return path


display = set()


class BaseCache(object):
    """
    Provides utilities for managing cached data files.

    Providers of caching functionality should derive from this, and appear
    as entries in ``registry``. The principle methods to override are
    ``_make_files()`` and ``_load()`` and ``_from_metadata()``.
    """
    # download block size in bytes
    blocksize = 5000000

    def __init__(self, driver, spec, catdir=None, cache_dir=None,
                 storage_options={}):
        """
        Parameters
        ----------
        driver: str
            Name of the plugin that can load catalog entry
        spec: list
            Specification for caching the data source.
        cache_dir: str or None
            Explicit location of cache root directory
        catdir: str or None
            Directory containing the catalog from which this spec was made
        """
        self._driver = driver
        self._spec = spec
        cd = cache_dir or conf['cache_dir']
        if cd == 'catdir':
            if catdir is None:
                raise TypeError('cache_dir="catdir" only allowed when loaded'
                                'from a catalog file.')
            cd = os.path.join(catdir, 'intake_cache')
        self._cache_dir = cd

        self._storage_options = storage_options
        self._metadata = CacheMetadata()
    
    def _ensure_cache_dir(self):
        if not os.path.exists(self._cache_dir):
            os.makedirs(self._cache_dir)
        if os.path.isfile(self._cache_dir):
            raise Exception("Path for cache directory exists as a file: {}"
                            "".format(self._cache_dir))

    def _munge_path(self, cache_subdir, urlpath):
        import re

        regex = sanitize_path(self._spec['regex'])
        path = sanitize_path(urlpath)

        cache_path = re.sub(
            r"%s" % regex,
            os.path.join(self._cache_dir, cache_subdir),
            path
        )

        return urlpath if path == cache_path else cache_path

    def _hash(self, urlpath):
        return md5(
                str((os.path.basename(urlpath), 
                     self._spec.get('regex', ''),
                     self._driver)).encode()
            ).hexdigest()

    def _path(self, urlpath, subdir=None):
        if subdir is None:
            subdir = self._hash(urlpath)
        cache_path = self._munge_path(subdir, urlpath)
        dirname = os.path.dirname(cache_path)
        if not os.path.exists(dirname):
            os.makedirs(dirname)

        return cache_path

    def _log_metadata(self, urlpath, original_path, cache_path):
        metadata = {
            'created': datetime.now().isoformat(),
            'original_path': original_path,
            'cache_path': cache_path
            }
        self._metadata.update(urlpath, metadata)

    def load(self, urlpath, output=None, **kwargs):
        """
        Downloads data from a given url, generates a hashed filename,
        logs metadata, and caches it locally.

        Parameters
        ----------

        urlpath: str, location of data
            May be a local path, or remote path if including a protocol specifier
            such as ``'s3://'``. May include glob wildcards.
        output: bool
            Whether to show progress bars; turn off for testing

        Returns
        -------
        List of local cache_paths to be opened instead of the remote file(s). If
        caching is disable, the urlpath is returned.
        """
        if conf.get('cache_disabled', False):
            return [urlpath]
        self.output = output if output is not None else conf.get(
            'cache_download_progress', True)

        cache_paths = self._from_metadata(urlpath)
        if cache_paths is None:
            files_in, files_out = self._make_files(urlpath)
            self._load(files_in, files_out, urlpath)
        cache_paths = self._from_metadata(urlpath)
        return cache_paths

    def _from_metadata(self, urlpath):
        """Return set of local URLs if files already exist"""
        md = self.get_metadata(urlpath)
        if md is not None:
            return [e['cache_path'] for e in md]

    def _load(self, files_in, files_out, urlpath, meta=True):
        """Download a set of files"""
        import dask
        out = []
        outnames = []
        for file_in, file_out in zip(files_in, files_out):
            cache_path = file_out.path
            outnames.append(cache_path)

            # If `_munge_path` did not find a match we want to avoid
            # writing to the urlpath.
            if cache_path == urlpath:
                continue

            if not os.path.isfile(cache_path):
                logger.debug("Caching file: {}".format(file_in.path))
                logger.debug("Original path: {}".format(urlpath))
                logger.debug("Cached at: {}".format(cache_path))
                if meta:
                    self._log_metadata(urlpath, file_in.path, cache_path)
                ddown = dask.delayed(_download)
                out.append(ddown(file_in, file_out, self.blocksize,
                                 self.output))
        dask.compute(*out)
        return outnames

    def _make_files(self, urlpath, **kwargs):
        """Make OpenFiles for all input/outputs"""
        raise NotImplementedError

    def get_metadata(self, urlpath):
        """
        Parameters
        ----------

        urlpath: str, location of data
            May be a local path, or remote path if including a protocol specifier
            such as ``'s3://'``. May include glob wildcards.

        Returns
        -------
        Metadata (dict) about a given urlpath.
        """
        return self._metadata.get(urlpath)
    
    def clear_cache(self, urlpath):
        """
        Clears cache and metadata for a given urlpath.

        Parameters
        ----------

        urlpath: str, location of data
            May be a local path, or remote path if including a protocol specifier
            such as ``'s3://'``. May include glob wildcards.
        """
        cache_entries = self._metadata.pop(urlpath, [])  # ignore if missing
        for cache_entry in cache_entries:
            try:
                os.remove(cache_entry['cache_path'])
            except (OSError, IOError):
                pass
        try:
            fn = os.path.dirname(cache_entry['cache_path'])
            os.rmdir(fn)
        except (OSError, IOError):
            logger.debug("Failed to remove cache directory: %s" % fn)
    
    def clear_all(self):
        """
        Clears all cache and metadata.
        """
        for urlpath in self._metadata.keys():
            self.clear_cache(urlpath)
        
        # Safely clean up anything else.
        if not os.path.isdir(self._cache_dir):
            return
        for subdir in os.listdir(self._cache_dir):
            try:
                fn = os.path.join(self._cache_dir, subdir)
                if os.path.isdir(fn):
                    shutil.rmtree(fn)
                if os.path.isfile(fn):
                    os.remove(fn)
            except (OSError, IOError) as e:
                logger.warning(str(e))


def _download(file_in, file_out, blocksize, output=False):
    """Read from input and write to output file in blocks"""
    with warnings.catch_warnings():
        warnings.filterwarnings('ignore')

        if output:
            try:
                from tqdm.autonotebook import tqdm
            except ImportError:
                logger.warn("Cache progress bar requires tqdm to be installed:"
                            " conda/pip install tqdm")
                output = False
        if output:
            try:
                file_size = file_in.fs.size(file_in.path)
                pbar_disabled = False
            except ValueError as err:
                logger.debug("File system error requesting size: {}".format(err))
                file_size = 0
                pbar_disabled = True
            for i in range(100):
                if i not in display:
                    display.add(i)
                    out = i
                    break
            pbar = tqdm(total=file_size // 2 ** 20, leave=False,
                        disable=pbar_disabled,
                        position=out, desc=os.path.basename(file_out.path),
                        mininterval=0.1,
                        bar_format=r'{n}/|/{l_bar}')

        logger.debug("Caching {}".format(file_in.path))
        with file_in as f1:
            with file_out as f2:
                data = True
                while data:
                    data = f1.read(blocksize)
                    f2.write(data)
                    if output:
                        pbar.update(len(data) // 2**20)
        if output:
            try:
                pbar.update(pbar.total - pbar.n)  # force to full
                pbar.close()
            except Exception as e:
                logger.debug('tqdm exception: %s' % e)
            finally:
                display.remove(out)


class FileCache(BaseCache):
    """Cache specific set of files

    Input is a single file URL, URL with glob characters or list of URLs. Output
    is a specific set of local files.
    """

    def _make_files(self, urlpath, **kwargs):
        from dask.bytes import open_files

        self._ensure_cache_dir()
        subdir = self._hash(urlpath)
        files_in = open_files(urlpath, 'rb', **self._storage_options)
        files_out = [open_files([self._path(f.path, subdir)], 'wb',
                                **self._storage_options)[0]
                     for f in files_in]
        return files_in, files_out


class DirCache(BaseCache):
    """Cache a complete directory tree

    Input is a directory root URL, plus a ``depth`` parameter for how many
    levels of subdirectories to search. All regular files will be copied. Output
    is the resultant local directory tree.
    """

    def _make_files(self, urlpath, **kwargs):
        from dask.bytes import open_files

        self._ensure_cache_dir()
        subdir = self._hash(urlpath)
        depth = self._spec['depth']
        files_in = []
        for i in range(1, depth + 1):
            files_in.extend(open_files('/'.join([urlpath] + ['*']*i)))
        files_out = [open_files([self._path(f.path, subdir)], 'wb',
                                **self._storage_options)[0]
                     for f in files_in]
        files_in2, files_out2 = [], []
        paths = set(os.path.dirname(f.path) for f in files_in)
        for fin, fout in zip(files_in, files_out):
            if fin.path in paths:
                try:
                    os.makedirs(fout.path)
                except Exception:
                    pass
            else:
                files_in2.append(fin)
                files_out2.append(fout)
        return files_in2, files_out2

    def _from_metadata(self, urlpath):
        """Return set of local URLs if files already exist"""
        md = self.get_metadata(urlpath)
        if md is not None:
            return [self._path(urlpath)]


class CompressedCache(BaseCache):
    """Download and decompress cacher

    For one or more remote compressed files, downloads to local temporary and
    extracts all contained files to local cache. Input is URL(s) (including
    globs) pointing to remote compressed files, plus optional ``decomp``,
    which is "infer" by default (guess from file extension) or one of the
    key strings in ``intake.source.decompress.decomp``. Output is the list
    of extracted files.
    """

    def _make_files(self, urlpath, **kwargs):
        import tempfile
        d = tempfile.mkdtemp()
        from dask.bytes import open_files

        self._ensure_cache_dir()
        self._urlpath = urlpath
        files_in = open_files(urlpath, 'rb')
        files_out = [open_files(
            [os.path.join(d, os.path.basename(f.path))], 'wb',
                                **self._storage_options)[0]
             for f in files_in]
        super(CompressedCache, self)._load(files_in, files_out, urlpath,
                                           meta=False)
        return files_in, files_out

    def _load(self, files_in, files_out, urlpath, meta=True):
        from .decompress import decomp
        subdir = self._path(urlpath)
        try:
            os.makedirs(subdir)
        except (OSError, IOError):
            pass
        files = [f.path for f in files_out]
        out = []
        for f, orig in zip(files, files_in):
            # TODO: add snappy, brotli, lzo, lz4, xz... ?
            if 'decomp' in self._spec and self._spec['decomp'] != 'infer':
                d = self._spec['decomp']
            elif f.endswith('.zip'):
                d = 'zip'
            elif f.endswith(".tar.gz") or f.endswith('.tgz'):
                d = 'tgz'
            elif f.endswith(".tar.bz2") or f.endswith('.tbz'):
                d = 'tbz'
            elif f.endswith(".tar"):
                d = 'tar'
            elif f.endswith('.gz'):
                d = 'gz'
            elif f.endswith('.bz2'):
                d = 'bz'
            if d not in decomp:
                raise ValueError('Unknown compression for "%s"' % f)
            out2 = decomp[d](f, subdir)
            for fn in out2:
                logger.debug("Caching file: {}".format(f))
                logger.debug("Original path: {}".format(orig.path))
                logger.debug("Cached at: {}".format(fn))
                if meta:
                    self._log_metadata(self._urlpath, orig.path, fn)
                out.append(fn)
        return out


class CacheMetadata(collections.MutableMapping):
    """
    Utility class for managing persistent metadata stored in the Intake config directory.
    """
    def __init__(self, *args, **kwargs):
        from intake import config

        self._path = os.path.join(config.confdir, 'cache_metadata.json')
        d = os.path.dirname(self._path)
        if not os.path.exists(d):
            os.makedirs(d)

        if os.path.isfile(self._path):
            with open(self._path) as f:
                self._metadata = json.load(f)
        else:
            self._metadata = {}

    def __getitem__(self, key):
        return self._metadata[self.__keytransform__(key)]

    def __setitem__(self, key, value):
        self._metadata[self.__keytransform__(key)] = value
        self._save()

    def __delitem__(self, key):
        del self._metadata[self.__keytransform__(key)]
        self._save()

    def __iter__(self):
        return iter(self._metadata)

    def __len__(self):
        return len(self._metadata)

    def __keytransform__(self, key):
        return key

    def update(self, key, cache_entry):
        entries = self._metadata.get(key, [])
        entries.append(cache_entry)
        self._metadata[key] = entries
        self._save()

    def _save(self):
        with open(self._path, 'w') as f:
            json.dump(self._metadata, f)

    def pop(self, key, default=None):
        item = self._metadata.pop(key, default)
        self._save()
        return item
    
    def keys(self):
        return list(self._metadata.keys())


registry = {
    'file': FileCache,
    'dir': DirCache,
    'compressed': CompressedCache
}


def make_caches(driver, specs, catdir=None, cache_dir=None, storage_options={}):
    """
    Creates Cache objects from the cache_specs provided in the catalog yaml file
    
    Parameters
    ----------
    
    driver: str
        Name of the plugin that can load catalog entry
    specs: list
        Specification for caching the data source.
    """
    if specs is None:
        return []
    return [registry.get(spec['type'], FileCache)(
        driver, spec, catdir=catdir, cache_dir=cache_dir,
        storage_options=storage_options)
        for spec in specs]



from intake.source.base import Schema
from .base import RemoteSource, get_partition


class RemoteDataFrame(RemoteSource):
    """Dataframe on an Intake server"""

    name = 'remote_dataframe'
    container = 'dataframe'

    def __init__(self, url, headers, **kwargs):
        super(RemoteDataFrame, self).__init__(url, headers, **kwargs)
        self.npartitions = kwargs['npartitions']
        self.shape = tuple(kwargs['shape'])
        self.metadata = kwargs['metadata']
        self.dtype = kwargs['dtype']
        self._schema = Schema(npartitions=self.npartitions,
                              extra_metadata=self.metadata,
                              dtype=self.dtype,
                              shape=self.shape,
                              datashape=None)
        self.dataframe = None

    def _load_metadata(self):
        import dask.dataframe as dd
        import dask.delayed
        if self.dataframe is None:
            self.parts = [dask.delayed(get_partition)(
                self.url, self.headers, self._source_id, self.container, i
            )
                          for i in range(self.npartitions)]
            self.dataframe = dd.from_delayed(self.parts)
        return self._schema

    def _get_partition(self, i):
        self._load_metadata()
        return self.parts[i].compute()

    def read(self):
        self._load_metadata()
        return self.dataframe.compute()

    def to_dask(self):
        self._load_metadata()
        return self.dataframe

    def _close(self):
        self.dataframe = None




from os.path import expanduser
import logging
import os
import yaml
logger = logging.getLogger('intake')

confdir = os.getenv('INTAKE_CONF_DIR',
                    os.path.join(expanduser('~'), '.intake'))
conffile = os.getenv('INTAKE_CONF_FILE', None)


defaults = {
    'auth': {'class': 'intake.auth.base.BaseAuth'},
    'port': 5000,
    'cache_dir': os.path.join(expanduser('~'), '.intake/cache'),
    'cache_disabled': False,
    'cache_download_progress': True,
    'logging': 'INFO',
    'catalog_path': []
    }
conf = {}


def reset_conf():
    """Set conf values back to defaults"""
    conf.clear()
    conf.update(defaults)


def cfile():
    return os.getenv('INTAKE_CONF_FILE', None) or os.path.join(
        confdir, 'conf.yaml')


def save_conf(fn=None):
    """Save current configuration to file as YAML

    If not given, uses current config directory, ``confdir``, which can be
    set by INTAKE_CONF_DIR.
    """
    if fn is None:
        fn = cfile()
    try:
        os.makedirs(os.path.dirname(fn))
    except (OSError, IOError):
        pass
    with open(fn, 'w') as f:
        yaml.dump(conf, f)


def load_conf(fn=None):
    """Update global config from YAML file

    If fn is None, looks in global config directory, which is either defined
    by the INTAKE_CONF_DIR env-var or is ~/.intake/ .
    """
    if fn is None:
        fn = cfile()
    if os.path.isfile(fn):
        with open(fn) as f:
            try:
                conf.update(yaml.load(f))
            except Exception as e:
                logger.warning('Failure to load config file "{fn}": {e}'
                               ''.format(fn=fn, e=e))


def intake_path_dirs(path):
    """Return a list of directories from the intake path."""
    separator = ';' if os.name == 'nt' else ':'
    return path.split(separator)


reset_conf()
load_conf(conffile)
# environment variables take precedence over conf file
if 'INTAKE_CACHE_DIR' in os.environ:
    conf['cache_dir'] = os.environ['INTAKE_CACHE_DIR']
if 'INTAKE_DISABLE_CACHING' in os.environ:
    conf['cache_disabled'] = os.environ['INTAKE_DISABLE_CACHING'
                                        ].lower() == 'true'
if 'INTAKE_CACHE_PROGRESS' in os.environ:
    conf['cache_download_progress'] = os.environ['INTAKE_CACHE_PROGRESS'
                                      ].lower() == 'true'
if 'INTAKE_LOG_LEVEL' in os.environ:
    conf['logging'] = os.environ['INTAKE_LOG_LEVEL']
if 'INTAKE_PATH' in os.environ:
    conf['catalog_path'] = intake_path_dirs(os.environ['INTAKE_PATH'])
logger.setLevel(conf['logging'])
ch = logging.StreamHandler()
formatter = logging.Formatter('%(asctime)s %(name)s:%(levelname)s, %(message)s')
ch.setFormatter(formatter)
logger.addHandler(ch)
logger.debug('Intake logger set to debug')



import msgpack
import requests
from intake.source.base import DataSource, Schema
from . import serializer
from ..compat import unpack_kwargs
from intake import __version__


class RemoteSource(DataSource):
    """Base class for all DataSources living on an Intake server"""
    version = __version__

    def __init__(self, url, headers, name, parameters, metadata=None, **kwargs):
        """

        Parameters
        ----------
        url: str
            Address of the server
        headers: dict
            HTTP headers to sue in calls
        name: str
            handle to reference this data
        parameters: dict
            To pass to the server when it instantiates the data source
        metadata: dict
            Additional info
        kwargs: ignored
        """
        super(RemoteSource, self).__init__(self)
        self.url = url
        self.name = name
        self.parameters = parameters
        self.headers = headers
        self._source_id = None
        self.metadata = metadata or {}
        self._get_source_id()

    def _get_source_id(self):
        if self._source_id is None:
            payload = dict(action='open', name=self.name,
                           parameters=self.parameters)
            req = requests.post(self.url, data=msgpack.packb(
                payload, use_bin_type=True), **self.headers)
            req.raise_for_status()
            response = msgpack.unpackb(req.content, **unpack_kwargs)
            self._parse_open_response(response)

    def _parse_open_response(self, response):
        self.datashape = response['datashape']
        dtype_descr = response['dtype']
        if isinstance(dtype_descr, list):
            # Reformat because NumPy needs list of tuples
            dtype_descr = [tuple(x) for x in response['dtype']]
        self.dtype = dtype_descr
        self.shape = tuple(response['shape'] or ())
        self.npartitions = response['npartitions']
        self.metadata = response['metadata']
        self._schema = Schema(datashape=None, dtype=self.dtype,
                              shape=self.shape,
                              npartitions=self.npartitions,
                              metadata=self.metadata)
        self._source_id = response['source_id']

    def _get_partition(self, i):
        raise NotImplementedError

    def to_dask(self):
        raise NotImplementedError


def get_partition(url, headers, source_id, container, partition):
    """Serializable function for fetching a data source partition

    Parameters
    ----------
    url: str
        Server address
    headers: dict
        HTTP header parameters
    source_id: str
        ID of the source in the server's cache (unique per user)
    container: str
        Type of data, like "dataframe" one of ``intake.container.container_map``
    partition: serializable
        Part of data to fetch, e.g., an integer for a dataframe.
    """
    accepted_formats = list(serializer.format_registry.keys())
    accepted_compression = list(serializer.compression_registry.keys())
    payload = dict(action='read',
                   source_id=source_id,
                   accepted_formats=accepted_formats,
                   accepted_compression=accepted_compression)

    if partition is not None:
        payload['partition'] = partition

    try:
        resp = requests.post(url, data=msgpack.packb(
            payload, use_bin_type=True), **headers)
        if resp.status_code != 200:
            raise Exception('Error reading data')

        msg = msgpack.unpackb(resp.content, **unpack_kwargs)
        format = msg['format']
        compression = msg['compression']
        compressor = serializer.compression_registry[compression]
        encoder = serializer.format_registry[format]
        chunk = encoder.decode(compressor.decompress(msg['data']),
                               container)
        return chunk
    finally:
        if resp is not None:
            resp.close()



import msgpack
import requests
import six

from ..source import registry as plugin_registry
from .entry import CatalogEntry
from .utils import expand_defaults, coerce


class RemoteCatalogEntry(CatalogEntry):
    """An entry referring to a remote data definition"""
    def __init__(self, url, auth, *args, **kwargs):
        """

        Parameters
        ----------
        url: str
            HTTP address of the Intake server this entry comes from
        auth: Auth instance
            If there are additional headers to add to calls, this instance will
            provide them
        kwargs: additional keys describing the entry, name, description,
            container,
        """
        self.url = url
        self.auth = auth
        self.args = args
        self.kwargs = kwargs
        self._description = self.kwargs.get('description', "")
        self._metadata = self.kwargs.get('metatata', {})
        getenv = kwargs.pop('getenv', True)
        getshell = kwargs.pop('getshell', True)
        self.http_args = kwargs.pop('http_args', {}).copy()
        if 'headers' not in self.http_args:
            self.http_args['headers'] = {}
        super(RemoteCatalogEntry, self).__init__(getenv=getenv,
                                                 getshell=getshell)

    def describe(self):
        return self.kwargs

    def describe_open(self, **kwargs):
        return {
            'plugin': None,
            'description': self._description,
            'direct_access': False,
            'metadata': self._metadata,
            'args': (self.url, )
        }

    def get(self, **user_parameters):
        for par in self.kwargs['user_parameters']:
            if par['name'] not in user_parameters:
                default = par['default']
                if isinstance(default, six.string_types):
                    default = coerce(par['type'], expand_defaults(
                        par['default'], True, self.getenv, self.getshell))
                user_parameters[par['name']] = default
        entry = self.kwargs

        http_args = self.http_args.copy()
        http_args['headers'] = self.http_args['headers'].copy()
        http_args['headers'].update(self.auth.get_headers())
        return open_remote(
            self.url, entry['name'], container=entry['container'],
            user_parameters=user_parameters, description=entry['description'],
            http_args=http_args
            )


def open_remote(url, entry, container, user_parameters, description, http_args):
    """Create either local direct data source or remote streamed source"""
    from intake.container import container_map
    if url.startswith('intake://'):
        url = url[len('intake://'):]
    payload = dict(action='open',
                   name=entry,
                   parameters=user_parameters,
                   available_plugins=list(plugin_registry.keys()))
    req = requests.post(url, data=msgpack.packb(
        payload, use_bin_type=True), **http_args)
    if req.ok:
        response = msgpack.unpackb(req.content, encoding='utf-8')

        if 'plugin' in response:
            # Direct access
            source = plugin_registry[response['plugin']](**response['args'])
        else:
            # Proxied access
            source = container_map[container](
                url, http_args, parameters=user_parameters,
                name=entry, **response)
        source.description = response.get('description', description)
        return source

    else:
        raise Exception('Server error: %d, %s' % (req.status_code, req.reason))



import os
import subprocess
import pytest


def test_reset(tempdir):
    subprocess.call('INTAKE_CONF_DIR=%s intake config reset' % tempdir,
                    shell=True)
    fn = os.path.join(tempdir, 'conf.yaml')
    assert os.path.isfile(fn)
    txt = open(fn).read()
    assert 'port: 5000' in txt


def test_info(tempdir):
    out = subprocess.check_output('INTAKE_CONF_DIR=%s intake config info' %
                                  tempdir, shell=True).decode()
    fn = os.path.join(tempdir, 'conf.yaml')
    assert fn in out
    assert 'INTAKE_CONF_DIR' in out
    assert 'INTAKE_CONF_FILE' not in out
    assert "(does not exist)" in out
    with open(fn, 'w') as f:
        f.write('port: 5000')
    out = subprocess.check_output('INTAKE_CONF_DIR=%s intake config info' %
                                  tempdir, shell=True).decode()
    assert "(does not exist)" not in out


def test_defaults(tempdir):
    out = subprocess.check_output('INTAKE_CONF_DIR=%s intake config list-'
                                  'defaults' % tempdir, shell=True).decode()
    assert 'port: 5000' in out


def test_get(tempdir):
    fn = os.path.join(tempdir, 'conf.yaml')
    with open(fn, 'w') as f:
        f.write('port: 5001')
    out = subprocess.check_output('INTAKE_CONF_DIR=%s intake config get' %
                                  tempdir, shell=True).decode()
    assert 'port: 5001' in out
    out = subprocess.check_output('INTAKE_CONF_DIR=%s intake config get port' %
                                  tempdir, shell=True).decode()
    assert out == '5001\n'


def test_log_level():
    out = subprocess.check_output('INTAKE_LOG_LEVEL=DEBUG intake config info',
                                  shell=True, stderr=subprocess.STDOUT).decode()
    assert "logger set to debug" in out




import importlib


def get_auth_class(auth, *args, **kwargs):
    """Instantiate class from a string spec and arguments

    Parameters
    ----------
    auth: str
        Something like ``package.module.AuthClass``
    args, kwargs: passed to the class's init function

    Returns
    -------
    Instance of the given class
    """
    mod, klass = auth.rsplit('.', 1)
    module = importlib.import_module(mod)
    cl = getattr(module, klass)
    return cl(*args, **kwargs)



import pytest


def verify_plugin_interface(plugin):
    assert isinstance(plugin.version, str)
    assert isinstance(plugin.container, str)
    assert isinstance(plugin.partition_access, bool)


def verify_datasource_interface(source):
    for attr in ['container', 'description', 'datashape', 'dtype', 'shape',
                 'npartitions', 'metadata']:
        assert hasattr(source, attr)

    for method in ['discover', 'read', 'read_chunked', 'read_partition',
                   'to_dask', 'close']:
        assert hasattr(source, method)


@pytest.fixture
def temp_cache(tempdir):
    import intake
    old = intake.config.conf.copy()
    olddir = intake.config.confdir
    intake.config.confdir = str(tempdir)
    intake.config.conf.update({'cache_dir': str(tempdir),
                               'cache_download_progress': False,
                               'cache_disabled': False})
    intake.config.save_conf()
    try:
        yield
    finally:
        intake.config.confdir = olddir
        intake.config.conf.update(old)



import functools
import itertools
from jinja2 import Environment, meta, Template
import os
import re
import shlex
import subprocess
import sys

import six


def flatten(iterable):
    """Flatten an arbitrarily deep list"""
    # likely not used
    iterable = iter(iterable)
    while True:
        try:
            item = next(iterable)
        except StopIteration:
            break

        if isinstance(item, six.string_types):
            yield item
            continue

        try:
            data = iter(item)
            iterable = itertools.chain(data, iterable)
        except:
            yield item


def reload_on_change(f):
    @functools.wraps(f)
    def wrapper(self, *args, **kwargs):
        self.reload()
        return f(self, *args, **kwargs)

    return wrapper


def clamp(value, lower=0, upper=sys.maxsize):
    """Clamp float between given range"""
    return max(lower, min(upper, value))


def _expand(p, context, all_vars):
    if isinstance(p, dict):
        return {k: _expand(v, context, all_vars) for k, v in p.items()}
    elif isinstance(p, (list, tuple, set)):
        return type(p)(_expand(v, context, all_vars) for v in p)
    elif isinstance(p, six.string_types):
        ast = Environment().parse(p)
        all_vars -= meta.find_undeclared_variables(ast)
        return Template(p).render(context)
    else:
        # no expansion
        return p


def expand_templates(pars, context, return_left=False):
    """
    Render variables in context into the set of parameters with jinja2.

    For variables that are not strings, nothing happens.

    Parameters
    ----------
    pars: dict
        values are strings containing some jinja2 controls
    context: dict
        values to use while rendering
    return_left: bool
        whether to return the set of variables in context that were not used
        in rendering parameters

    Returns
    -------
    dict with the same keys as ``pars``, but updated values; optionally also
    return set of unused parameter names.
    """
    all_vars = set(context)
    out = _expand(pars, context, all_vars)
    if return_left:
        return out, all_vars
    return out


def expand_defaults(default, client=False, getenv=True, getshell=True):
    """Compile env, client_env, shell and client_shell commands

    Execution rules:
    - env() and shell() execute where the cat is loaded, if getenv and getshell
      are True, respectively
    - client_env() and client_shell() execute only if client is True and
      getenv/getshell are also True.

    If both getenv and getshell are False, this method does nothing.

    If the environment variable is missing or the shell command fails, the
    output is an empty string.
    """
    r = re.match(r'env\((.*)\)', default)
    if r and not client and getenv:
        default = os.environ.get(r.groups()[0], '')
    r = re.match(r'client_env\((.*)\)', default)
    if r and client and getenv:
        default = os.environ.get(r.groups()[0], '')
    r = re.match(r'shell\((.*)\)', default)
    if r and not client and getshell:
        try:
            cmd = shlex.split(r.groups()[0])
            default = subprocess.check_output(
                cmd).rstrip().decode('utf8')
        except (subprocess.CalledProcessError, OSError):
            default = ''
    r = re.match(r'client_shell\((.*)\)', default)
    if r and client and getshell:
        try:
            cmd = shlex.split(r.groups()[0])
            default = subprocess.check_output(
                cmd).rstrip().decode('utf8')
        except (subprocess.CalledProcessError, OSError):
            default = ''
    return default


def coerce_datetime(v=None):
    import pandas
    return pandas.to_datetime(v) if v else pandas.to_datetime(0)


COERCION_RULES = {
    'bool': bool,
    'datetime': coerce_datetime,
    'float': float,
    'int': int,
    'list': list,
    'str': str,
    'unicode': six.text_type
}


def coerce(dtype, value):
    """
    Convert a value to a specific type.

    If the value is already the given type, then the original value is
    returned. If the value is None, then the default value given by the
    type constructor is returned. Otherwise, the type constructor converts
    and returns the value.
    """
    if type(value).__name__ == dtype:
        return value
    op = COERCION_RULES[dtype]
    return op() if value is None else op(value)



import os
import os.path
import time

from tornado.ioloop import IOLoop
from tornado.testing import AsyncHTTPTestCase
import msgpack

from intake import Catalog
from intake.container.serializer import MsgPackSerializer, GzipCompressor
from intake.cli.server.server import IntakeServer
from intake.compat import unpack_kwargs


class TestServerV1Base(AsyncHTTPTestCase):
    def get_app(self):
        catalog_file = os.path.join(os.path.dirname(__file__), 'catalog1.yml')
        local_catalog = Catalog(catalog_file)
        self.server = IntakeServer(local_catalog)
        return self.server.make_app()

    def encode(self, msg):
        return msgpack.packb(msg, use_bin_type=True)

    def decode(self, bytestr):
        return msgpack.unpackb(bytestr, **unpack_kwargs)


class TestServerV1Info(TestServerV1Base):
    def test_info(self):
        response = self.fetch('/v1/info')
        self.assertEqual(response.code, 200)

        info = self.decode(response.body)

        self.assert_('version' in info)

        expected = [
            {
                'container': 'dataframe',
                'direct_access': 'forbid',
                'description': 'example1 source plugin',
                'name': 'use_example1',
                'user_parameters': []},
            {
                'container': 'dataframe',
                'direct_access': 'forbid',
                'description': 'entry1 full',
                'name': 'entry1',
                'user_parameters': []},
            {
                'container': 'dataframe',
                'direct_access': 'allow',
                'description': 'entry1 part',
                'name': 'entry1_part',
                'user_parameters': [{
                    'allowed': ['1', '2'],
                    'default': '1',
                    'description': 'part of filename',
                    'name': 'part',
                    'type': 'str'
                }]
            }
        ]

        def sort_by_name(seq):
            return sorted(seq, key=lambda d: d['name'])

        for left, right in zip(sort_by_name(info['sources']),
                               sort_by_name(expected)):
            self.assertDictEqual(left, right)


class TestServerV1Source(TestServerV1Base):
    def make_post_request(self, msg, expected_status=200):
        request = self.encode(msg)
        response = self.fetch('/v1/source', method='POST', body=request,
                              headers={'Content-type':
                                       'application/vnd.msgpack'})
        self.assertEqual(response.code, expected_status)

        responses = []
        unpacker = msgpack.Unpacker(**unpack_kwargs)
        unpacker.feed(response.body)

        for msg in unpacker:
            responses.append(msg)

        return responses

    def test_open(self):
        import pickle
        msg = dict(action='open', name='entry1', parameters={})
        resp_msg, = self.make_post_request(msg)

        self.assertEqual(resp_msg['container'], 'dataframe')
        self.assertEqual(resp_msg['shape'], [None, 3])
        expected_dtype = {'name': 'object', 'score': 'float64', 'rank': 'int64'}
        actual_dtype = resp_msg['dtype']
        self.assertEqual(expected_dtype, actual_dtype)
        self.assertEqual(resp_msg['npartitions'], 2)
        md = resp_msg['metadata']
        md.pop('catalog_dir', None)
        self.assertEqual(md, dict(foo='bar', bar=[1, 2, 3], cache=[]))

        self.assert_(isinstance(resp_msg['source_id'], str))

    def test_open_direct(self):
        msg = dict(action='open', name='entry1_part', parameters=dict(part='2'),
                   available_plugins=['csv'])
        resp_msg, = self.make_post_request(msg)

        self.assert_('csv' in resp_msg['plugin'])
        args = resp_msg['args']
        self.assertEquals(set(args.keys()), set(['urlpath', 'metadata']))
        self.assert_(args['urlpath'].endswith('/entry1_2.csv'))
        md = args['metadata']
        md.pop('catalog_dir', None)
        self.assertEquals(md, dict(foo='baz', bar=[2, 4, 6], cache=[]))
        self.assertEqual(resp_msg['description'], 'entry1 part')

    def test_read_part_compressed(self):
        msg = dict(action='open', name='entry1', parameters={})
        resp_msg, = self.make_post_request(msg)
        source_id = resp_msg['source_id']

        msg2 = dict(action='read', source_id=source_id,
                    accepted_formats=['msgpack'], accepted_compression=['gzip'],
                    partition=0)
        resp_msgs = self.make_post_request(msg2)

        self.assertEqual(len(resp_msgs), 1)
        ser = MsgPackSerializer()
        comp = GzipCompressor()

        for chunk in resp_msgs:
            self.assertEqual(chunk['format'], 'msgpack')
            self.assertEqual(chunk['compression'], 'gzip')
            self.assertEqual(chunk['container'], 'dataframe')

            data = ser.decode(comp.decompress(chunk['data']),
                              container='dataframe')
            self.assertEqual(len(data), 4)

    def test_read_partition(self):
        msg = dict(action='open', name='entry1', parameters={})
        resp_msg, = self.make_post_request(msg)
        source_id = resp_msg['source_id']

        msg2 = dict(action='read', partition=1, source_id=source_id,
                    accepted_formats=['msgpack'])
        resp_msgs = self.make_post_request(msg2)

        self.assertEqual(len(resp_msgs), 1)
        ser = MsgPackSerializer()

        part = resp_msgs[0]
        self.assertEqual(part['format'], 'msgpack')
        self.assertEqual(part['compression'], 'none')
        self.assertEqual(part['container'], 'dataframe')

        data = ser.decode(part['data'], container='dataframe')
        self.assertEqual(len(data), 4)

    def test_bad_action(self):
        msg = dict(action='bad', name='entry1')
        response, = self.make_post_request(msg, expected_status=400)
        self.assertIn('bad', response['error'])

    def test_no_format(self):
        msg = dict(action='open', name='entry1', parameters={})
        resp_msg, = self.make_post_request(msg)
        source_id = resp_msg['source_id']

        msg2 = dict(action='read', source_id=source_id,
                    accepted_formats=['unknown_format'])
        response, = self.make_post_request(msg2, expected_status=400)
        self.assertIn('compatible', response['error'])

    def test_idle_timer(self):
        self.server.start_periodic_functions(close_idle_after=0.1,
                                             remove_idle_after=0.2)

        msg = dict(action='open', name='entry1', parameters={})
        resp_msg, = self.make_post_request(msg)
        source_id = resp_msg['source_id']

        # Let ioloop run once with do-nothing function to make sure source
        # isn't closed
        time.sleep(0.05)
        IOLoop.current().run_sync(lambda: None)

        # Cheat and look into internal state now
        source = self.server._cache.peek(source_id)
        assert source._dataframe is not None

        # now wait slightly over idle time, run periodic functions,
        # and check again
        time.sleep(0.06)
        IOLoop.current().run_sync(lambda: None)

        # should be closed
        source = self.server._cache.peek(source_id)
        assert source._dataframe is None

        # wait a little longer
        time.sleep(0.1)
        IOLoop.current().run_sync(lambda: None)

        # source should be gone
        with self.assertRaises(KeyError):
            self.server._cache.peek(source_id)



import os
import shutil
import time
import tempfile

import intake
import appdirs
import pytest


def copy_test_file(filename, target_dir):
    if not os.path.exists(target_dir):
        os.makedirs(target_dir)  # can't use exist_ok in Python 2.7

    # Put a catalog file in the user catalog directory
    test_dir = os.path.dirname(__file__)
    test_catalog = os.path.join(test_dir, filename)
    target_catalog = os.path.join(target_dir, '__unit_test_'+filename)

    shutil.copyfile(test_catalog, target_catalog)
    return target_catalog


@pytest.fixture
def user_catalog():
    target_catalog = copy_test_file('catalog1.yml',
                                    appdirs.user_data_dir(appname='intake',
                                                          appauthor='intake'))
    yield target_catalog
    # Remove the file, but not the directory (because there might be other
    # files already there)
    os.remove(target_catalog)

@pytest.fixture
def tmp_path_catalog():
    tmp_path = os.path.join(tempfile.gettempdir(), 'intake')
    try:
        os.makedirs(tmp_path)
    except:
        pass
    target_catalog = copy_test_file('catalog1.yml', tmp_path)
    yield target_catalog
    # Remove the file, but not the directory (because there might be other
    # files already there)
    os.remove(target_catalog)

def test_autoregister_open():
    assert hasattr(intake, 'open_csv')


def test_default_catalogs():
    # No assumptions about contents of these catalogs.
    # Just make sure they exist and don't raise exceptions
    list(intake.cat)


def test_user_catalog(user_catalog):
    cat = intake.load_combo_catalog()
    time.sleep(2) # wait 2 seconds for catalog to refresh
    assert set(cat) >= set(['ex1', 'ex2'])


def test_path_catalog(tmp_path_catalog):
    intake.config.conf['catalog_path'] = [os.path.join(tempfile.gettempdir(), 'intake')]
    cat = intake.load_combo_catalog()
    time.sleep(2) # wait 2 seconds for catalog to refresh
    assert set(cat) >= set(['ex1', 'ex2'])
    del intake.config.conf['catalog_path']


def test_bad_open():
    with pytest.raises(ValueError):
        intake.open_catalog(None, driver='unknown')


def test_output_notebook():
    pytest.importorskip('hvplot')
    intake.output_notebook()



from __future__ import print_function

import sys
import os
import argparse
import shutil
import yaml

from intake import Catalog


def print_entry_info(catalog, name):
    info = catalog[name].describe()
    for key in sorted(info.keys()):
        print("[{}] {}={}".format(name, key, info[key]))


def listing(args):
    catalog = Catalog(args.uri)
    for entry in list(catalog):
        if args.full:
            print_entry_info(catalog, entry)
        else:
            print(entry)


def precache(args):
    catalog = Catalog(args.uri)
    for entry in list(catalog):
        s = catalog[entry]()
        if s.cache:
            print("Caching for entry %s" % entry)
            s.read()


def describe(args):
    catalog = Catalog(args.uri)
    print_entry_info(catalog, args.name)


def exists(args):
    catalog = Catalog(args.uri)
    print(args.name in catalog)


def get(args):
    catalog = Catalog(args.uri)
    with catalog[args.name].get() as f:
        print(f.read())


def discover(args):
    catalog = Catalog(args.uri)
    with catalog[args.name].get() as f:
        print(f.discover())


def example(args):
    print('Creating example catalog...')
    files = ['us_states.yml', 'states_1.csv', 'states_2.csv']
    for filename in files:
        if os.path.exists(filename):
            print('Cannot create example catalog in current directory.\n'
                  '%s already exists.' % filename)
            return 1

    src_dir = os.path.join(os.path.dirname(__file__), '..', 'sample')

    for filename in files:
        src_name = os.path.join(src_dir, filename)
        dest_name = filename
        dest_dir = os.path.dirname(filename)
        print('  Writing %s' % filename)
        if dest_dir != '' and not os.path.exists(dest_dir):
            os.mkdir(dest_dir)
        shutil.copyfile(src_name, dest_name)

    print('''\nTo load the catalog:
>>> import intake
>>> cat = intake.open_catalog('%s')
''' % files[0])


def conf_def(args):
    from intake.config import defaults
    print(yaml.dump(defaults, default_flow_style=False))


def conf_reset_save(args):
    from intake.config import reset_conf, save_conf
    reset_conf()
    save_conf()


def conf_get_key(args):
    from intake.config import conf
    if args.key:
        print(conf[args.key])
    else:
        print(yaml.dump(conf, default_flow_style=False))


def conf_show_info(args):
    from intake.config import cfile
    if 'INTAKE_CONF_DIR' in os.environ:
        print('INTAKE_CONF_DIR: ', os.environ['INTAKE_CONF_DIR'])
    if 'INTAKE_CONF_FILE' in os.environ:
        print('INTAKE_CONF_FILE: ', os.environ['INTAKE_CONF_FILE'])
    ex = "" if os.path.isfile(cfile()) else "(does not exist)"
    print('Using: ', cfile(), ex)


def cache_clear(args):
    from intake.source.cache import BaseCache
    c = BaseCache(None, None)
    if args.key is None:
        c.clear_all()
    else:
        c.clear_cache(args.key)


def cache_usage(args):
    from intake.config import conf
    total_size = 0
    for dirpath, dirnames, filenames in os.walk(conf['cache_dir']):
        for f in filenames:
            fp = os.path.join(dirpath, f)
            total_size += os.path.getsize(fp)
    for unit in ['', 'k', 'M', 'G', 'T', 'P', 'E', 'Z']:
        # "human"
        # https://gist.github.com/cbwar/d2dfbc19b140bd599daccbe0fe925597
        if abs(total_size) < 1024.0:
            s = "%3.1f %s" % (total_size, unit)
            break
        total_size /= 1024.0
    print("%s: %s" % (conf['cache_dir'], s))


def cache_list_files(args):
    from intake.source.cache import CacheMetadata
    md = CacheMetadata()
    print(yaml.dump(md[args.key], default_flow_style=False))


def cache_list_keys(args):
    from intake.source.cache import CacheMetadata
    md = CacheMetadata()
    print(yaml.dump(list(md), default_flow_style=False))


def cache_clear_all(args):
    from intake.source.cache import BaseCache
    if args.key is None:
        BaseCache(None, None).clear_all()
    else:
        BaseCache(None, None).clear_cache(args.key)


def main(argv=None):
    if argv is None:
        argv = sys.argv

    parser = argparse.ArgumentParser(description='Intake Catalog CLI', add_help=False)
    subparsers = parser.add_subparsers(help='sub-command help')

    list_parser = subparsers.add_parser('list', help='catalog listing')
    list_parser.add_argument('--full', action='store_true')
    list_parser.add_argument('uri', metavar='URI', type=str, help='Catalog URI')
    list_parser.set_defaults(func=listing)

    describe_parser = subparsers.add_parser('describe', help='description for catalog entry')
    describe_parser.add_argument('uri', metavar='URI', type=str, help='Catalog URI')
    describe_parser.add_argument('name', metavar='NAME', type=str, help='Catalog name')
    describe_parser.set_defaults(func=describe)

    exists_parser = subparsers.add_parser('exists', help='existence check for catalog entry')
    exists_parser.add_argument('uri', metavar='URI', type=str, help='Catalog URI')
    exists_parser.add_argument('name', metavar='NAME', type=str, help='Catalog name')
    exists_parser.set_defaults(func=exists)

    precache_parser = subparsers.add_parser('precache', help='For each entry in a catalog which defined caching,'
                                                             ' populate the caching.')
    precache_parser.add_argument('uri', metavar='URI', type=str, help='Catalog URI')
    precache_parser.set_defaults(func=precache)

    get_parser = subparsers.add_parser('get', help='get catalog entry')
    get_parser.add_argument('uri', metavar='URI', type=str, help='Catalog URI')
    get_parser.add_argument('name', metavar='NAME', type=str, help='Catalog name')
    get_parser.set_defaults(func=get)

    discover_parser = subparsers.add_parser('discover', help='discover catalog entry')
    discover_parser.add_argument('uri', metavar='URI', type=str, help='Catalog URI')
    discover_parser.add_argument('name', metavar='NAME', type=str, help='Catalog name')
    discover_parser.set_defaults(func=discover)

    example_parser = subparsers.add_parser('example', help='create example catalog')
    example_parser.set_defaults(func=example)

    conf_parser = subparsers.add_parser('config', help='configuration functions')
    conf_sub = conf_parser.add_subparsers()
    conf_list = conf_sub.add_parser('list-defaults', help='Show all builtin defaults')
    conf_list.set_defaults(func=conf_def)
    conf_reset = conf_sub.add_parser('reset', help='Set config file to defaults')
    conf_reset.set_defaults(func=conf_reset_save)
    conf_info = conf_sub.add_parser('info', help='Show config settings')
    conf_info.set_defaults(func=conf_show_info)
    conf_get = conf_sub.add_parser('get', help='Get current config, specific key or all')
    conf_get.add_argument('key', type=str, help='Key in config dictionary',
                          nargs='?')
    conf_get.set_defaults(func=conf_get_key)

    cache_parser = subparsers.add_parser('cache', help='Locally cached files')
    cache_sub = cache_parser.add_subparsers()
    cache_list = cache_sub.add_parser('list-keys', help='List keys currently stored')
    cache_list.set_defaults(func=cache_list_keys)
    cache_files = cache_sub.add_parser('list-files', help='List files for a give key')
    cache_files.add_argument('key', type=str, help='Key to list files for')
    cache_files.set_defaults(func=cache_list_files)
    cache_rm = cache_sub.add_parser('clear')
    cache_rm.add_argument('key', type=str, help='Key to remove (all, if omitted)',
                          nargs='?')
    cache_rm.set_defaults(func=cache_clear_all)
    cache_du = cache_sub.add_parser('usage')
    cache_du.set_defaults(func=cache_usage)

    if not argv[1:]:
        parser.print_usage()
        sys.exit(1)

    args = parser.parse_args()
    if not hasattr(args, 'func') and '-h' not in argv:
        argv.append('-h')
        args = parser.parse_args()
    retcode = args.func(args)
    if retcode is None:
        retcode = 0
    return retcode


if __name__ == "__main__":
    sys.exit(main(sys.argv))



from __future__ import print_function

import time
from uuid import uuid4

import logging
import msgpack
import pickle
import tornado.gen
import tornado.ioloop
import tornado.web

from intake.config import conf
from intake.container import serializer
from intake.auth import get_auth_class
from intake import __version__
logger = logging.getLogger('intake')


class IntakeServer(object):
    """Main intake-server tornado application"""
    def __init__(self, catalog):
        self._catalog = catalog
        self._cache = SourceCache()
        self._periodic_callbacks = []
        auth = conf.get('auth', 'intake.auth.base.BaseAuth')
        logger.debug('auth: %s' % auth)
        self._auth = get_auth_class(auth['class'], *auth.get('args', tuple()),
                                    **auth.get('kwargs', {}))

    def get_handlers(self):
        return [
            (r"/v1/info", ServerInfoHandler,
             dict(catalog=self._catalog, cache=self._cache, auth=self._auth)),
            (r"/v1/source", ServerSourceHandler,
             dict(catalog=self._catalog, cache=self._cache, auth=self._auth)),
        ]

    def make_app(self):
        handlers = self.get_handlers()
        return tornado.web.Application(handlers)

    def start_periodic_functions(self, close_idle_after=None,
                                 remove_idle_after=None):
        if len(self._periodic_callbacks) > 0:
            raise Exception('Periodic functions already started '
                            'for this server')

        # Disabling periodic timers with None makes testing easier
        if close_idle_after is not None:
            cache_closer = self._make_cache_closer(close_idle_after)
            self._periodic_callbacks.append(cache_closer)

        if remove_idle_after is not None:
            cache_remover = self._make_cache_remover(remove_idle_after)
            self._periodic_callbacks.append(cache_remover)

        for callback in self._periodic_callbacks:
            callback.start()

    def _make_cache_closer(self, idle_time):
        def cache_closer_callback():
            self._cache.close_idle(idle_time)
        return self._make_cache_callback(cache_closer_callback, idle_time)

    def _make_cache_remover(self, idle_time):
        def cache_remover_callback():
            self._cache.remove_idle(idle_time)
        return self._make_cache_callback(cache_remover_callback, idle_time)

    def _make_cache_callback(self, callback, idle_time):
        # Check ever 1/10 of the idle_time
        interval_ms = (idle_time / 10.0) * 1000
        callback = tornado.ioloop.PeriodicCallback(callback, interval_ms)
        return callback


class ServerInfoHandler(tornado.web.RequestHandler):
    """Basic info about the server"""
    def initialize(self, cache, catalog, auth):
        self.cache = cache
        self.catalog = catalog
        self.auth = auth

    def get(self):
        head = self.request.headers
        if self.auth.allow_connect(head):
            if 'source_id' in head:
                cat = self.cache.get(head['source_id'])
            else:
                cat = self.catalog
            sources = []
            for name, source in cat.walk(depth=1).items():
                if self.auth.allow_access(head, source, self.catalog):
                    info = source.describe()
                    info['name'] = name
                    sources.append(info)

            server_info = dict(version=__version__, sources=sources,
                               metadata=self.catalog.metadata)
        else:
            msg = 'Access forbidden'
            raise tornado.web.HTTPError(status_code=403, log_message=msg,
                                        reason=msg)
        self.write(msgpack.packb(server_info, use_bin_type=True))


class SourceCache(object):
    """Stores DataSources requested by some user"""
    def __init__(self):
        self._sources = {}

    def add(self, source):
        source_id = str(uuid4())
        now = time.time()
        self._sources[source_id] = dict(source=source, open_time=now,
                                        last_time=now)
        logger.debug('Adding %s to cache, uuid %s' % (source, source_id))
        return source_id

    def get(self, uuid):
        record = self._sources[uuid]
        record['last_time'] = time.time()
        return record['source']

    def peek(self, uuid):
        """Get the source but do not change the last access time"""
        return self._sources[uuid]['source']

    def touch(self, uuid):
        record = self._sources[uuid]
        record['last_time'] = time.time()

    def close_idle(self, idle_secs):
        threshold = time.time() - idle_secs

        for record in self._sources.values():
            if record['last_time'] < threshold:
                record['source'].close()

    def remove_idle(self, idle_secs):
        threshold = time.time() - idle_secs

        # Make a copy of the items so we can mutate the dictionary
        for uuid, record in list(self._sources.items()):
            if record['last_time'] < threshold:
                logger.debug('Removing source %s from cache' % uuid)
                del self._sources[uuid]


class ServerSourceHandler(tornado.web.RequestHandler):
    """Open or stream data source

    The requests "action" field (open|read) specified what the request wants
    to do. Open caches the source and created an ID for it, read uses that
    ID to reference the source and read a partition.
    """
    def initialize(self, catalog, cache, auth):
        self._catalog = catalog
        self._cache = cache
        self.auth = auth

    @tornado.gen.coroutine
    def post(self):
        request = msgpack.unpackb(self.request.body, encoding='utf-8')
        action = request['action']
        head = self.request.headers
        logger.debug('Source POST: %s' % request)

        if action == 'open':
            if 'source_id' in head:
                cat = self._cache.get(head['source_id'])
            else:
                cat = self._catalog
            entry_name = request['name']
            entry = cat[entry_name]
            if not self.auth.allow_access(head, entry, cat):
                msg = 'Access forbidden'
                raise tornado.web.HTTPError(status_code=403, log_message=msg,
                                            reason=msg)
            user_parameters = request['parameters']
            client_plugins = request.get('available_plugins', [])

            # Can the client directly access the data themselves?
            open_desc = entry.describe_open(**user_parameters)
            direct_access = open_desc['direct_access']
            plugin_name = open_desc['plugin']
            client_has_plugin = plugin_name in client_plugins

            if direct_access == 'forbid' or \
                    (direct_access == 'allow' and not client_has_plugin):
                logger.debug("Opening entry %s" % entry)
                source = entry.get(**user_parameters)
                try:
                    source.on_server = True
                    source.discover()
                except Exception as e:
                    raise tornado.web.HTTPError(status_code=400,
                                                log_message="Discover failed",
                                                reason=str(e))
                source_id = self._cache.add(source)
                logger.debug('Container: %s, ID: %s' % (source.container,
                                                        source_id))
                response = dict(source._schema)
                response.update(dict(container=source.container,
                                     source_id=source_id,
                                     metadata=source.metadata))
                self.write(msgpack.packb(response, use_bin_type=True))
                self.finish()
            elif direct_access == 'force' and not client_has_plugin:
                msg = 'client must have plugin "%s" to access source "%s"' \
                      '' % (plugin_name, entry_name)
                raise tornado.web.HTTPError(status_code=400, log_message=msg,
                                            reason=msg)
            else:
                # If we get here, the client can access the source directly
                response = dict(
                    plugin=plugin_name, args=open_desc['args'],
                    description=open_desc['description'])
                self.write(msgpack.packb(response, use_bin_type=True))
                self.finish()

        elif action == 'read':
            source_id = request['source_id']
            source = self._cache.get(source_id)
            accepted_formats = request['accepted_formats']
            accepted_compression = request.get('accepted_compression', ['none'])
            partition = request.get('partition', None)

            chunk_encoder = self._pick_encoder(accepted_formats,
                                                     accepted_compression,
                                                     source.container)

            logger.debug("Read partition %s" % partition)
            if partition is not None:
                chunk = source.read_partition(partition)
            else:
                assert source.npartitions < 2
                chunk = source.read()

            data = chunk_encoder.encode(chunk, source.container)
            msg = dict(format=chunk_encoder.format_name,
                       compression=chunk_encoder.compressor_name,
                       container=source.container, data=data)
            self.write(msgpack.packb(msg, use_bin_type=True))
            self.flush()
            self._cache.touch(source_id)  # keep source alive

            self.finish()

        else:
            msg = '"%s" not a valid source action' % action
            raise tornado.web.HTTPError(status_code=400, log_message=msg,
                                        reason=msg)

    def _pick_encoder(self, accepted_formats, accepted_compression, container):
        format_encoder = None

        for f in accepted_formats:
            if f in serializer.format_registry:
                format_encoder = serializer.format_registry[f]

        if format_encoder is None:
            msg = 'Unable to find compatible format'
            raise tornado.web.HTTPError(status_code=400, log_message=msg,
                                        reason=msg)

        compressor = serializer.NoneCompressor()  # Default
        for f in accepted_compression:
            if f in serializer.compression_registry:
                compressor = serializer.compression_registry[f]

        return serializer.ComboSerializer(format_encoder, compressor)

    def write_error(self, status_code, **kwargs):
        error_exception = kwargs.get('exc_info', None)
        if error_exception is not None:
            msg = dict(error=str(error_exception[1]))
        else:
            msg = dict(error='unknown error')
        self.write(msgpack.packb(msg, use_bin_type=True))
